"""

Helm workload Metta plugin

As a workload, using a kube_api client, manage helm charts

"""
from typing import Any, List, Dict
import logging
import subprocess
import shutil
from enum import Enum

import yaml

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import Fixtures
from mirantis.testing.metta.client import METTA_PLUGIN_INTERFACE_ROLE_CLIENT
from mirantis.testing.metta_health.healthcheck import Health

from .kubeapi_client import METTA_PLUGIN_ID_KUBERNETES_CLIENT

logger = logging.getLogger("metta.contrib.kubernetes.workload.helm")

METTA_PLUGIN_ID_KUBERNETES_HELM_WORKLOAD = "metta_kubernetes_helm_workload"
""" workload plugin_id for the metta_kubernetes helm plugin """


KUBERNETES_HELM_WORKLOAD_CONFIG_LABEL = "kubernetes"
""" default config label used to load the workload """
KUBERNETES_HELM_WORKLOAD_CONFIG_BASE = "workload.helm"
""" default config key for creating a workload object """

KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VERSION = "version"
""" config key for helm version """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_REPOS = "repos"
""" config key for helm repos dict which need to be added """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_CHART = "chart"
""" config key for helm chart, either local path or http(s) """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUESSET = "set"
""" config key for helm chart values """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUESFILE_VALUES = "values"
""" config key for helm chart values that should be put into a values file """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUESFILE_PATH = "file"
""" config key for helm chart path to values file that we should create """
KUBERNETES_HELM_WORKLOAD_CONFIG_DEFAULT_VALUESFILE_PATH = "values.yml"
""" config default value for helm chart path to values file that we should create """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_NAMESPACE = "namespace"
""" config key for namespace to install to """

KUBERNETES_HELM_WORKLOAD_DEFAULT_NAMESPACE = "default"
""" default namespace to install to if no namespace was passed """

KUBERNETES_HELM_WORKLOAD_DEFAULT_BIN = "helm"
""" default helm executble path """
KUBERNETES_HELM_WORKLOAD_DEFAULT_WORKINGDIR = "."
""" default helm working dir for subprocess """


class Status(Enum):
    """A Helm Status enum."""

    UNKNOWN = "unknown"
    DEPLOYED = "deployed"
    UNINSTALLED = "uninstalled"
    SUPERSEDED = "superseded"
    FAILED = "failed"
    UNINSTALLING = "uninstalling"
    PENDING_INSTALL = "pending-install"
    PENDING_UPGRADE = "pending-upgrade"
    PENDING_ROLLBACK = "pending-rollback"


# This is effectively a struct for interpreting the release status information
# pylint: disable=too-few-public-methods, too-many-instance-attributes
class HelmReleaseStatus:
    """Interpreted helm release status.

    Used to formalize the response object from a status request

    """

    def __init__(self, status_response: dict):
        """Interpret status from a status response."""
        self.name = status_response["name"]
        self.version = status_response["version"]
        self.namespace = status_response["namespace"]

        # self.first_deployed = datetime.strptime(status_response['info']['first_deployed'],
        #                                         '%Y-%m-%dT%H:%M:%S.%fZ')
        # self.last_deployed = datetime.strptime(status_response['info']['last_deployed'],
        #                                         '%Y-%m-%dT%H:%M:%S.%fZ')

        self.deleted = status_response["info"]["deleted"]
        self.description = status_response["info"]["description"]
        self.status = Status(status_response["info"]["status"])
        self.notes = status_response["info"]["notes"] if "notes" in status_response["info"] else ""

        self.config = status_response["config"]

        self.manifest = status_response["manifest"]


# pylint: disable=too-many-instance-attributes
class KubernetesHelmWorkloadPlugin:
    """Kubernetes workload class."""

    # this is what we need for this purpose
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        environment,
        instance_id,
        label: str = KUBERNETES_HELM_WORKLOAD_CONFIG_LABEL,
        base: Any = KUBERNETES_HELM_WORKLOAD_CONFIG_BASE,
        helm_bin: str = KUBERNETES_HELM_WORKLOAD_DEFAULT_BIN,
        work_dir: str = KUBERNETES_HELM_WORKLOAD_DEFAULT_WORKINGDIR,
    ):
        """Run the super constructor but also set class properties.

        This implements the args part of the client interface.

        Here we expect to receiv config pointers so that we can determine what
        helm workload to apply, and what values will be needed.

        Parameters:
        -----------
        file (str) : path to where we should put values yaml file

        bin (str) : helm executable path
        dir (dir) : working dir to be used with subprocess

        """
        self._environment: Environment = environment
        """ Environemnt in which this plugin exists """
        self._instance_id: str = instance_id
        """ Unique id for this plugin instance """

        self._config_label: str = label
        """ configerus load label that should contain all of the config """
        self._config_base: str = base
        """ configerus get key that should contain all tf config """

        self._kubeconfig: str = ""
        """Path to a kubeconfig file to be used by helm (see .prepare())."""

        workload_config = self._environment.config().load(self._config_label)

        self.namespace: str = workload_config.get(
            [self._config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_NAMESPACE],
            default=KUBERNETES_HELM_WORKLOAD_DEFAULT_NAMESPACE,
        )
        """String k8s namespace to constrain created resources."""

        self._chart: str = workload_config.get(
            [self._config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_CHART]
        )
        """Helm chart to use for the release."""

        self.set: Dict[str, str] = workload_config.get(
            [self._config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUESSET],
            default={},
        )
        """Dict of --set values."""
        self.values: Dict[str, Any] = workload_config.get(
            [self._config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUESFILE_VALUES],
            default={},
        )
        """Nested Dict of values to be written to a values file."""
        self._file: str = workload_config.get(
            [self._config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUESFILE_PATH],
            default=KUBERNETES_HELM_WORKLOAD_CONFIG_DEFAULT_VALUESFILE_PATH,
        )
        """String path to use for the values file."""

        self._repos: Dict[str, str] = workload_config.get(
            [self._config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_REPOS], default={}
        )
        """Helm repos to be added."""

        self._working_dir: str = work_dir
        """Path to the helm chart, which is used as a python subprocess chdir."""

        if shutil.which(helm_bin) is None:
            raise ValueError(
                "Helm binary not found. Helm commands cannot be called. "
                f"Expected binary at path {helm_bin}"
            )

        self._bin: str = helm_bin
        """Path to the helm binary."""

        # do an initial prepare in case it is never properly run
        try:
            self.prepare()
        # pylint: disable=broad-except
        except Exception:
            pass

    # deep argument is an info() standard across plugins
    # pylint: disable=unused-argument
    def info(self, deep: bool = False):
        """Return dict data about this plugin for introspection."""
        return {
            "release": {
                "name": self._instance_id,
                "repos": self._repos,
                "chart": self._chart,
                "set": self.set,
                "values": self.values,
            },
            "required_fixtures": {
                "kubernetes": {
                    "interface": METTA_PLUGIN_INTERFACE_ROLE_CLIENT,
                    "plugin_id": METTA_PLUGIN_ID_KUBERNETES_CLIENT,
                },
            },
        }

    def health(self) -> Health:
        """Create a Health check for the helm workload."""
        health: Health = Health(source=self._instance_id)

        status = self.status()

        if status.status in [Status.UNKNOWN]:
            health.unknown(
                f"Helm: {self._instance_id} release status is unknown: {status.status} "
                f": {status.description}"
            )

        if status.status in [Status.SUPERSEDED, Status.UNINSTALLING]:
            health.warning(
                f"Helm: {self._instance_id} release status is at issue: {status.status} "
                f": {status.description}"
            )

        if status.status in [Status.DEPLOYED, Status.UNINSTALLED]:
            health.healthy(
                f"Helm: {self._instance_id} release status is good: {status.status} "
                f": {status.description}"
            )

        if status.status == Status.FAILED:
            health.error(
                f"Helm: {self._instance_id} release status is not good: {status.status} "
                f": {status.description}"
            )

        if status.status in [
            Status.PENDING_INSTALL,
            Status.PENDING_UPGRADE,
            Status.PENDING_ROLLBACK,
        ]:
            health.warning(f"Helm status pending: {status.status}")

        return health

    def prepare(self, fixtures: Fixtures = None):
        """Create a workload instance from a set of fixtures.

        Parameters:
        -----------
        fixtures (Fixtures) : a set of fixtures that this workload will use to
            retrieve a kubernetes client plugin.

        """
        if fixtures is None:
            fixtures = self._environment.fixtures()

        try:
            client = fixtures.get_plugin(
                plugin_id=METTA_PLUGIN_ID_KUBERNETES_CLIENT,
            )
        except KeyError as err:
            raise NotImplementedError(
                "Workload could not find the needed client: "
                f"{METTA_PLUGIN_ID_KUBERNETES_CLIENT}"
            ) from err

        self._kubeconfig = client.config_file

    def apply(self, wait: bool = True, debug: bool = False):
        """Apply the helm chart.

        To make the helm apply method reusable, we always run an upgrade with
        the --install flag, which w orks for both the first install, and any
        upgrades run after install

        You need to run the prepare() method before this one.

        Parameters:
        -----------
        wait (bool) : ask the helm client to wait until resources are created
            before returning.

        debug (bool) : ask the helm client for verbose output

        """
        for repo_name, repo_url in self._repos.items():
            self._run(cmd=["repo", "add", repo_name, repo_url])

        cmd = ["upgrade"]
        cmd += [self._instance_id, self._chart]

        cmd += ["--install", "--create-namespace"]

        if wait:
            cmd += ["--wait"]
        if debug:
            cmd += ["--debug"]

        if len(self.set):
            # turn the set dict into '--set a=A,b=B,c=C'
            cmd += [
                "--set",
                ",".join([f"{name}={value}" for (name, value) in self.values.items()]),
            ]

        if len(self.values):
            # turn the values into a file, and add it to the command
            with open(self._file, "w", encoding="utf8") as val_file:
                yaml.dump(self.values, val_file)

            cmd += ["--values", self._file]

        try:
            self._run(cmd=cmd)
        except Exception as err:
            raise RuntimeError("Helm failed to install the release") from err

    # -all is the used command flag, so the var name makes sense
    # pylint: disable=redefined-builtin
    def list(
        self,
        all: bool = False,
        failed: bool = False,
        deployed: bool = False,
        pending: bool = False,
    ):
        """List all releases.

        This is not instance specific but still useful.

        """
        cmd = ["list", "--output=yaml"]

        if all:
            cmd += ["--all"]
        elif deployed:
            cmd += ["--deployed"]
        elif failed:
            cmd += ["--failed"]
        elif pending:
            cmd += ["--pending"]

        list_str = self._run(cmd=cmd, return_output=True)
        if list_str:
            return yaml.safe_load(list_str)

        return []

    def destroy(self, debug: bool = False):
        """Remove an installed helm release.

        Parameters:
        -----------
        debug (bool) : ask the helm client for verbose output

        """
        cmd = ["uninstall", self._instance_id]

        if debug:
            cmd += ["--debug"]

        self._run(cmd=cmd)

    def test(self):
        """Test an installed helm release.

        This runs the helm client test command.

        """
        self._run(cmd=["test", self._instance_id])

    def status(self) -> HelmReleaseStatus:
        """Get status of the installed helm release."""
        try:
            return HelmReleaseStatus(
                yaml.safe_load(
                    self._run(
                        cmd=["status", self._instance_id, "--output=yaml"],
                        return_output=True,
                    )
                )
            )

        except (subprocess.CalledProcessError, AttributeError) as err:
            return HelmReleaseStatus(
                {
                    "name": "unknown",
                    "version": "unknown",
                    "namespace": "unknown",
                    "info": {
                        "deleted": "unknown",
                        "description": str(err),
                        "status": "unknown",
                    },
                    "notes": "Status not found",
                    "config": {},
                    "manifest": [],
                }
            )

    def _run(self, cmd: List[str], return_output: bool = False):
        """Run a helm v3 command."""
        cmd = [
            self._bin,
            f"--kubeconfig={self._kubeconfig}",
            f"--namespace={self.namespace}",
        ] + cmd

        # this syntax makes it easier to read
        # pylint: disable=no-else-return
        if return_output:
            logger.debug("running launchpad command with output capture: %s", " ".join(cmd))
            return_exec: subprocess.CompletedProcess[bytes] = subprocess.run(
                cmd,
                cwd=self._working_dir,
                shell=False,
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            return_exec.check_returncode()
            return return_exec.stdout.decode("utf-8")
        else:
            logger.debug("running launchpad command: %s", " ".join(cmd))
            exec: subprocess.CompletedProcess[str] = subprocess.run(
                cmd, cwd=self._working_dir, check=True, text=True
            )
            exec.check_returncode()
            return exec



"""

Dummy metta provisioner plugin

A provisioner which doesn't do anything, but can still be configured to produce
various clients and outputs.  The provisioner has no significant requirements
and no impact.  Provisioning can be repeated or interupted without impact.

The dummy provisioner is entirely config based.  Use the prepare() method to
indicate an appropriate config source and the provisioner will do the rest. As
long as you match its config convention it can take care of itself.

"""

import logging
from typing import Dict, Any

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import Fixtures

logger = logging.getLogger("metta.contrib.dummy.provisioner")


class DummyProvisionerPlugin:
    """Dummy provisioner class"""

    def __init__(
        self,
        environment: Environment,
        instance_id: str,
        fixtures: Dict[str, Dict[str, Any]] = None,
    ):
        self._environment: Environment = environment
        """ Environemnt in which this plugin exists """
        self._instance_id: str = instance_id
        """ Unique id for this plugin instance """

        self.fixtures = Fixtures()
        """This plugin keeps fixtures."""
        if fixtures is not None:
            for child_instance_id, child_instance_dict in fixtures.items():
                child = environment.add_fixture_from_dict(
                    instance_id=child_instance_id, plugin_dict=child_instance_dict
                )
                self.fixtures.add(child)

    def apply(self):
        """pretend to bring a cluster up"""
        logger.info("%s:execute: apply()", self._instance_id)

    def prepare(self):
        """pretend to prepare the cluster"""
        logger.info("%s:execute: apply()", self._instance_id)

    def destroy(self):
        """pretend to brind a cluster down"""
        logger.info("%s:execute: apply()", self._instance_id)



"""

Metta terraform CLI plugin.

Provides functionality to inspect terraform configuration and execute
terraform operations, as well as check out terraform outputs.

"""
import logging
from typing import Dict, Any

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta_cli.base import CliBase, cli_output

from .provisioner import METTA_TERRAFORM_PROVISIONER_PLUGIN_ID
from .client import METTA_TERRAFORM_CLIENT_PLUGIN_ID

logger = logging.getLogger("metta.cli.terraform")

METTA_TERRAFORM_CLI_PLUGIN_ID = "metta_terraform_cli"
""" cli plugin_id for the info plugin """


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class TerraformCliPlugin(CliBase):
    """Fire command/group generator for terraform commands."""

    def fire(self):
        """Return a dict of commands.

        Don't return any commands if there are no plugind available

        """
        commands: Dict[str, object] = {}

        if (
            self._environment.fixtures().get(
                plugin_id=METTA_TERRAFORM_CLIENT_PLUGIN_ID,
                exception_if_missing=False,
            )
            is not None
        ):
            commands["terraform"] = TerraformClientGroup(self._environment)

        return commands


class TerraformClientGroup:
    """Terraform client cli commands."""

    def __init__(self, environment: Environment):
        """Inject environment."""
        self._environment: Environment = environment

        if (
            self._environment.fixtures().get(
                plugin_id=METTA_TERRAFORM_PROVISIONER_PLUGIN_ID,
                exception_if_missing=False,
            )
            is not None
        ):
            self.provisioner = TerraformProvisionerGroup(self._environment)

    def _select_client(self, instance_id: str = ""):
        """Pick a matching terraform provisioner."""
        if instance_id:
            return self._environment.fixtures().get(
                plugin_id=METTA_TERRAFORM_CLIENT_PLUGIN_ID,
                instance_id=instance_id,
            )

        # Get the highest priority provisioner
        return self._environment.fixtures().get(
            plugin_id=METTA_TERRAFORM_CLIENT_PLUGIN_ID,
        )

    def info(self, client: str = "", deep: bool = False) -> Dict[str, Any]:
        """Get info about a client plugin."""
        fixture = self._select_client(instance_id=client)
        return cli_output(fixture.info(deep=deep))

    def init(self, client: str = ""):
        """Run terraform init."""
        plugin = self._select_client(instance_id=client).plugin
        plugin.init()

    def plan(self, client: str = ""):
        """Run client plan."""
        plugin = self._select_client(instance_id=client).plugin
        plugin.plan()

    def apply(self, client: str = "", nolock: bool = False):
        """Run terraform apply."""
        plugin = self._select_client(instance_id=client).plugin
        plugin.apply(lock=(not nolock))

    def destroy(self, client: str = "", nolock: bool = False):
        """Run terraform destroy."""
        plugin = self._select_client(instance_id=client).plugin
        plugin.destroy(lock=(not nolock))

    def check(self, client: str = ""):
        """Run client check."""
        plugin = self._select_client(instance_id=client).plugin
        plugin.check()

    def state(self, client: str = ""):
        """Run client check."""
        plugin = self._select_client(instance_id=client).plugin
        return cli_output(plugin.state())

    def output(self, client: str = "", name: str = ""):
        """Retrieve Terraform outputs."""
        plugin = self._select_client(instance_id=client).plugin
        return cli_output(plugin.output(name=name))


class TerraformProvisionerGroup:
    """Base Fire command group for terraform cli commands."""

    def __init__(self, environment: Environment):
        """Inject environment."""
        self._environment: Environment = environment

    def _select_provisioner(self, instance_id: str = ""):
        """Pick a matching terraform provisioner."""
        if instance_id:
            return self._environment.fixtures().get(
                plugin_id=METTA_TERRAFORM_PROVISIONER_PLUGIN_ID,
                instance_id=instance_id,
            )

        # Get the highest priority provisioner
        return self._environment.fixtures().get(
            plugin_id=METTA_TERRAFORM_PROVISIONER_PLUGIN_ID,
        )

    def info(self, provisioner: str = "", deep: bool = False) -> Dict[str, Any]:
        """Get info about a provisioner plugin."""
        fixture = self._select_provisioner(instance_id=provisioner)
        return cli_output(fixture.info(deep=deep))

    def prepare(self, provisioner: str = ""):
        """Run provisioner prepare."""
        plugin = self._select_provisioner(instance_id=provisioner).plugin
        plugin.prepare()

    def apply(self, provisioner: str = "", nolock: bool = False):
        """Run provisioner apply."""
        plugin = self._select_provisioner(instance_id=provisioner).plugin
        plugin.apply(lock=(not nolock))

    def destroy(self, provisioner: str = "", nolock: bool = False):
        """Run provisioner destroy."""
        plugin = self._select_provisioner(instance_id=provisioner).plugin
        plugin.destroy(lock=(not nolock))



"""

METTA CLI setuptools entrypoint

Keep the entrypint minimal, and rely on .base for real work.

"""

import logging

import fire

from .root import Root

logger = logging.getLogger("metta.cli.entrypoint")


def main():
    """Main entrypoint"""
    fire.Fire(Root)


if __name__ == "__main__":
    main()



"""

Dummy client plugin

"""

import logging
from typing import Dict, Any

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import Fixtures

logger = logging.getLogger("metta.contrib.dummy.client")


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class DummyClientPlugin:
    """Dummy client class

    As with all dummies, this is a failsafe plugin, that should never throw any
    exceptions if used according to metta standards.

    It can be used as a placeholder during development, or it can be used to
    log client events and output for greater development and debugging.

    The client will log any method call, including unknown methods, and so it
    can be used in place of any client, if you don't need the methods to return
    anything
    """

    def __init__(
        self,
        environment: Environment,
        instance_id: str,
        fixtures: Dict[str, Dict[str, Any]] = None,
    ):
        """Sset class properties

        Arguments:
        ----------

        environment (Environment) : Environment in which thisplugin exists.

        instance_id (str) : unique identifier for this plugin instance.

        fixtures (dict) : You can pass in some fixture definitions which this
            class will turn into fixtures and make retrievable.  This is a big
            part of the dummy.

        """
        self._environment: Environment = environment
        """ Environemnt in which this plugin exists """
        self._instance_id: str = instance_id
        """ Unique id for this plugin instance """

        self.fixtures = Fixtures()
        """This plugin keeps fixtures."""
        if fixtures is not None:
            for child_instance_id, child_instance_dict in fixtures.items():
                child = environment.add_fixture_from_dict(
                    instance_id=child_instance_id, plugin_dict=child_instance_dict
                )
                self.fixtures.add(child)



"""

Docker workload plugin to run docker run commands, given docker host config.

This type of docker workload is simple to use but perhaps not
very feature rich.

"""
from typing import Any
import logging

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import Fixtures

from .client import DockerPyClientPlugin, METTA_PLUGIN_ID_DOCKER_CLIENT

logger = logging.getLogger("metta_docker.workload.run")

METTA_PLUGIN_ID_DOCKER_RUN_WORKLOAD = "metta_docker_run_workload"
""" metta plugin_id for the launchpad cli plugin """

DOCKER_RUN_WORKLOAD_CONFIG_LABEL = "docker"
""" Configerus label for retrieving docker run workloads """
DOCKER_RUN_WORKLOAD_CONFIG_BASE = "workload.run"
""" Configerus get base for retrieving the default run workload """

DOCKER_RUN_WORKLOAD_CONFIG_KEY_RUN = "run"
""" Configerus get base for the docker run arguments e.g. image: """


class DockerPyRunWorkloadPlugin:
    """Docker Run workload class for the DockerPy."""

    def __init__(
        self,
        environment: Environment,
        instance_id: str,
        label: str = DOCKER_RUN_WORKLOAD_CONFIG_LABEL,
        base: Any = DOCKER_RUN_WORKLOAD_CONFIG_BASE,
    ):
        """Run the super constructor but also set class properties.

        Parameters:
        -----------
        outputs (Dict[Dict]) : pass in a dictionary which defines outputs that
            should be returned

        clients (Dict[Dict]) : pass in a dictionary which defines which clients
            should be requested when working on a provisioner

        """
        self._environment: Environment = environment
        """ Environemnt in which this plugin exists """
        self._instance_id: str = instance_id
        """ Unique id for this plugin instance """

        self._config_label = label
        """ configerus load label that should contain all of the config """
        self._config_base = base
        """ configerus get key that should contain all tf config """

        run_config = self._environment.config().load(self._config_label)
        """Configerus LoadedConfig for the docker run label."""

        self._run_settings = run_config.get(
            [self._config_base, DOCKER_RUN_WORKLOAD_CONFIG_KEY_RUN]
        )
        """Arguments for docker run."""

        self._docker_client: DockerPyClientPlugin = None
        """Metta Docker client plugin."""

        # do an initial prepare in case it is never properly run
        try:
            self.prepare()
        # pylint: disable=broad-except
        except Exception:
            pass

    # deep argument is an info() standard across plugins
    # pylint: disable=unused-argument
    def info(self, deep: bool = False):
        """Return dict data about this plugin for introspection."""
        return {
            "workload": {
                "run": self._run_settings,
                "required_fixtures": {
                    "docker": {
                        "plugin_id": [METTA_PLUGIN_ID_DOCKER_CLIENT],
                    }
                },
            }
        }

    def prepare(self, fixtures: Fixtures = None):
        """Create a workload instance from a set of fixtures.

        Parameters:
        -----------
        fixtures (Fixtures) : a set of fixtures that this workload will use to
            retrieve a docker client plugin.

        """
        if fixtures is None:
            fixtures = self._environment.fixtures()

        self._docker_client = fixtures.get_plugin(interfaces=[METTA_PLUGIN_ID_DOCKER_CLIENT])

    def apply(self):
        """Run the workload.

        @NOTE Needs a docker client fixture to run.  Use .set_fixtures() first

        """
        assert "image" in self._run_settings, "Run command had no image"

        return self._docker_client.containers.run(**self._run_settings)

    # pylint: disable=fixme, no-self-use
    # TODO write the docker container rm
    def destroy(self):
        """Destroy a created docker run."""
        logger.warning("docker run workload destroy functionality is @TODO.")



"""

PLUGIN: Client plugins.

Base definition for client type plugins.  This is not very "pythonish" but it
gives us a chance to define a base class and some constants related to that
kind of plugin.

"""

import logging

logger = logging.getLogger("metta.client")

METTA_PLUGIN_INTERFACE_ROLE_CLIENT = "client"
""" metta plugin interface identifier for client plugins """

METTA_CLIENT_CONFIG_CLIENTS_LABEL = "clients"
""" A centralized configerus label for multiple clients """
METTA_CLIENT_CONFIG_CLIENT_LABEL = "client"
""" A centralized configerus label for a client """
METTA_CLIENT_CONFIG_CLIENTS_KEY = "clients"
""" A centralized configerus key for multiple clients """
METTA_CLIENT_CONFIG_CLIENT_KEY = "client"
""" A centralized configerus key for one client """



"""

PLUGINS: OUTPUT.

Output plugins define returned content from an operation that can be interacted
with in a predictable manner.  This gives some abstraction to allow coupling
between unknown plugins for passing data or messages.
While the output plugins themselves define their interface, meaning that you
have to know the type of plugin you are consuming, it does allow for a small
number of common intermediate representations along with allowing for an
extensible system.

It is expected that a small number of output plugins can handle most needs.

"""

METTA_PLUGIN_INTERFACE_ROLE_OUTPUT = "output"
""" metta plugin interface identifier for output plugins """

METTA_OUTPUT_CONFIG_OUTPUTS_LABEL = "outputs"
""" A centralized configerus load label for multiple outputs """
METTA_OUTPUT_CONFIG_OUTPUT_LABEL = "output"
""" A centralized configerus load label for an output """
METTA_OUTPUT_CONFIG_OUTPUTS_KEY = "outputs"
""" A centralized configerus key for multiple outputs """
METTA_OUTPUT_CONFIG_OUTPUT_KEY = "output"
""" A centralized configerus key for one output """



"""

Metta CLI : Config commands.

Various commands that allow introspection of the configerus config setup and
values.  There are commands for inspecting which plugins are avaialble, and
what configuration sources are setup, but also for examining rendered config
and formatting strings using config - all to allow verification of what
config is feeding metta.

"""
import logging

from configerus.loaded import LOADED_KEY_ROOT
from configerus.plugin import Type as ConfigerusType
from configerus.contrib.files import PLUGIN_ID_SOURCE_PATH
from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT

from mirantis.testing.metta.environment import Environment

from .base import CliBase, cli_output

logger = logging.getLogger("metta.cli.config")


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class ConfigCliPlugin(CliBase):
    """Fire command/group generator for config commands."""

    def fire(self):
        """Return a dict of commands."""
        return {"config": ConfigGroup(self._environment)}


class ConfigGroup:
    """Base Fire command group for output commands."""

    def __init__(self, environment: Environment):
        """Store environment in object."""
        self._environment: Environment = environment

    def plugins(self, plugin_id: str = "", instance_id: str = "", plugin_type: str = ""):
        """List configerus plugins."""
        configerus_plugin_list = []
        for instance in self._environment.config().plugins.get_instances(
            plugin_id=plugin_id, instance_id=instance_id, type=plugin_type
        ):
            configerus_plugin_list.append(
                {
                    "type": instance.type,
                    "plugin_id": instance.plugin_id,
                    "instance_id": instance.instance_id,
                    "priority": instance.priority,
                }
            )
        return cli_output(configerus_plugin_list)

    def sources(self, plugin_id: str = "", instance_id: str = "", deep: bool = False):
        """List configerus sources."""
        source_list = []
        for instance in self._environment.config().plugins.get_instances(
            plugin_id=plugin_id, instance_id=instance_id, type=ConfigerusType.SOURCE
        ):
            source = {
                "plugin_id": instance.plugin_id,
                "instance_id": instance.instance_id,
                "priority": instance.priority,
            }

            if deep:
                if instance.plugin_id == PLUGIN_ID_SOURCE_PATH:
                    source["path"] = instance.plugin.path
                if instance.plugin_id == PLUGIN_ID_SOURCE_DICT:
                    source["data"] = instance.plugin.data

            source_list.append(source)

        return cli_output(source_list)

    def loaded(self, raw: bool = False):
        """List loaded config labels."""
        loaded = self._environment.config().loaded
        value = list(loaded)
        if raw:
            return value
        return cli_output(value)

    def get(self, label: str, key: str = LOADED_KEY_ROOT):
        """Retrieve configuration from the config object.

        USAGE:

            metta config get {label} [{key}]

        """
        try:
            loaded = self._environment.config().load(label)
        except KeyError:
            return f"Could not find the config label '{label}'"

        try:
            value = loaded.get(key)
        except KeyError:
            return f"Could not find the config key '{key}'"

        return cli_output(value)

    # broad exception catch just to format error message
    # pylint: disable=broad-except
    def format(self, data: str, default_label: str = None, raw: bool = False):
        """Format a target string using config templating."""
        try:
            if default_label is None:
                default_label = "you did not specify a default"
            value = self._environment.config().format(data=data, default_label=default_label)
        except Exception as err:
            return f"Error occured in formatting: '{err}'"

        if raw:
            return value
        return cli_output(value)



"""

A provisioner that combines multiple provisioners.

The combo provisioner keeps an ordered collection of "backend" provisioner
plugins which it proxies for any provisioner operations.

"""
import logging
from typing import Any, List

from configerus.loaded import LOADED_KEY_ROOT
from configerus.contrib.jsonschema.validate import (
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL,
)
from configerus.validator import ValidationError

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import (
    Fixtures,
    Fixture,
    METTA_FIXTURE_CONFIG_KEY_PRIORITY,
)
from mirantis.testing.metta.plugin import METTA_PLUGIN_CONFIG_KEY_INSTANCEID
from mirantis.testing.metta.provisioner import (
    METTA_PLUGIN_INTERFACE_ROLE_PROVISIONER,
)

logger = logging.getLogger("metta.contrib.provisioner:combo")

METTA_PLUGIN_ID_PROVISIONER_COMBO = "combo_provisioner"
""" provisioner plugin_id for the combo plugin """

COMBO_PROVISIONER_CONFIG_LABEL = "provisioner"
""" Configerus label for loading config to set up this provisioner plugin """
COMBO_PROVISIONER_CONFIG_BACKENDS_KEY = "backends"
""" Config key for backends list """

COMBO_PROVISIONER_VALIDATE_JSONSCHEMA = {
    "type": "object",
    "properties": {"backends": {"type": "array", "items": {"$ref": "#/$defs/backend"}}},
    "$defs": {
        "backend": {
            "type": "object",
            "required": ["instance_id"],
            "properties": {
                "instance_id": {
                    "type": "string",
                    "description": "Backend provisioner instance_id. A provisioner instance with "
                    "this id  must exist in the environment.",
                },
                "priority": {
                    "type": "boolean",
                    "description": "Backend provisioner priority in the combo list.  If not "
                    "provided then The provisioner's priority will be used.",
                },
            },
        }
    },
    "required": ["backends"],
}
""" Validation jsonschema for terraform config contents """
COMBO_PROVISIONER_VALIDATE_TARGET = {
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL: COMBO_PROVISIONER_VALIDATE_JSONSCHEMA
}


class ComboProvisionerPlugin:
    """Combo Provisioner plugin class.

    This provisioner plugin is configered with a list of backends, which it
    will iterate across for every provisioner method call.  The backends have a
    priority which define the order of their call and every provisioner method
    will follow that order (or reverse it.)

    """

    def __init__(
        self,
        environment: Environment,
        instance_id: str,
        label: str = COMBO_PROVISIONER_CONFIG_LABEL,
        base: Any = LOADED_KEY_ROOT,
    ):
        """Run the super constructor but also set class properties.

        Interpret provided config and configure the object with all of the
        needed pieces for executing terraform commands

        Parameters:
        -----------
        environment (Environment) : All metta plugins receive the environment
            object in which they were created.
        instance_id (str) : all metta plugins receive their own string identity.

        label (str) : Configerus load label for finding plugin config.
        base (str) : Configerus get base key in which the plugin should look for
            all config.

        """
        self._environment: Environment = environment
        """Environemnt in which this plugin exists."""
        self._instance_id: str = instance_id
        """Unique id for this plugin instance."""

        try:
            combo_config = self._environment.config().load(label)
        except KeyError as err:
            raise ValueError("Combo provisioner could not find any configurations") from err

        # Run confgerus validation on the config using our above defined jsonschema
        try:
            combo_config.get(base, validator=COMBO_PROVISIONER_VALIDATE_TARGET)
        except ValidationError as err:
            raise ValueError("Combo provisioner config failed validation") from err

        try:
            backends_list = combo_config.get([base, COMBO_PROVISIONER_CONFIG_BACKENDS_KEY])
            if not isinstance(backends_list, list):
                raise ValueError(
                    "Combo provisioner could not understand the backend list."
                    " A list was expected."
                )
        except KeyError as err:
            raise ValueError("Combo provisioner received no backend list from config.") from err

        # for each of our string instance_ids we add the backend in order by finding if from the
        # environment and adding it to our UCCTFixturesPlugin fixtures list.
        self.backends = Fixtures()
        for backend in backends_list:
            backend_instance_id = backend[METTA_PLUGIN_CONFIG_KEY_INSTANCEID]
            try:
                fixture = self._environment.fixtures().get(
                    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_PROVISIONER],
                    instance_id=backend_instance_id,
                )
            except KeyError as err:
                raise ValueError(
                    "Combo provisioner was given a backend provisioner key that it could not "
                    f"correlate with a existing fixture: {backend_instance_id}"
                ) from err

            if hasattr(backend, METTA_FIXTURE_CONFIG_KEY_PRIORITY):
                fixture.priority = backend[METTA_FIXTURE_CONFIG_KEY_PRIORITY]

            self.backends.add(fixture)

    def _get_backend_iter(self, low_to_high: bool = False):
        """Get the sorted backend fixtures.

        Parameters:
        -----------
        low-to-high (bool) : ask for the fixtures in a lowest to highest
            (reverse) order.

        Returns:
        --------
        Iterator which is either the backends fixtures object, or the backends
            reversed()

        """
        if low_to_high:
            return reversed(self.backends)
        return self.backends

    def info(self, deep: bool = False):
        """Return structured data about self."""
        backends_info = []
        # List backends in high->low priority as this shows the order of apply
        for backend in self.backends:
            backends_info.append(backend.info(deep=deep))

        return {"backends": backends_info}

    def prepare(self):
        """Prepare the provisioner to apply resources."""
        for backend_fixture in self._get_backend_iter():
            logger.info(
                "--> running backend prepare: [High->Low] %s",
                backend_fixture.instance_id,
            )
            backend_fixture.plugin.prepare()

    def apply(self):
        """Bring a cluster to the configured state."""
        for backend_fixture in self._get_backend_iter():
            logger.info("--> running backend apply: [High->Low] %s", backend_fixture.instance_id)
            backend_fixture.plugin.apply()

    def destroy(self):
        """Remove all resources created for the cluster."""
        for backend_fixture in self._get_backend_iter(low_to_high=True):
            logger.info(
                "--> running backend destroy: [Low->High] %s",
                backend_fixture.instance_id,
            )
            backend_fixture.plugin.destroy()

    # --- Fixture management
    #
    # We duplicate the UCCTFixturesPlugin methods, despite using it as a parent,
    # so that we can identify as that object, but because we need to allow all
    # backends to participate in fixture definition in order of priority.
    #
    # We of course have to override the any method which depends on our ordered
    # backend retrievals of get_fixtures() so that it doesn't run the parent
    # get_fixtures.

    def get_fixtures(
        self, instance_id: str = "", plugin_id: str = "", interfaces: List[str] = None
    ) -> Fixtures:
        """Retrieve any matching fixtures from any of the backends."""
        matches = Fixtures()
        for backend_fixture in self._get_backend_iter():
            plugin = backend_fixture.plugin
            if hasattr(plugin, "fixtures"):
                matches.merge(
                    plugin.fixtures.filter(
                        plugin_id=plugin_id,
                        interfaces=interfaces,
                        instance_id=instance_id,
                    )
                )
        return matches

    def get_fixture(
        self,
        plugin_id: str = "",
        interfaces: List[str] = None,
        instance_id: str = "",
        exception_if_missing: bool = True,
    ) -> Fixture:
        """Retrieve the first matching fixture from ordered backends."""
        matches = self.get_fixtures(
            plugin_id=plugin_id,
            instance_id=instance_id,
            interfaces=interfaces,
        )

        if len(matches) > 0:
            return matches.get()

        if exception_if_missing:
            raise KeyError("No matching fixture was found")
        return None

    def get_plugin(
        self,
        plugin_id: str = "",
        interfaces: List[str] = None,
        instance_id: str = "",
        exception_if_missing: bool = True,
    ) -> object:
        """Retrieve one of the backend fixtures."""
        fixture = self.get_fixture(
            plugin_id=plugin_id,
            interfaces=interfaces,
            instance_id=instance_id,
            exception_if_missing=exception_if_missing,
        )

        if fixture is not None:
            return fixture.plugin

        # this if is not needed, as get_fixture() handles exception_if_missing
        if exception_if_missing:
            raise KeyError("No matching plugin was found")
        return None



"""

Metta CLI : Provisioner commands.

Various commands that allow introspection of provisioner plugins/fixtures and
their contents.

"""
import logging


from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.provisioner import METTA_PLUGIN_INTERFACE_ROLE_PROVISIONER

from .base import CliBase, cli_output

logger = logging.getLogger("metta.cli.provisioner")


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class ProvisionerCliPlugin(CliBase):
    """Fire command/group generator for provisioner commands."""

    def fire(self):
        """Return a dict of commands."""
        return {"provisioner": ProvisionerGroup(self._environment)}


class ProvisionerGroup:
    """Base Fire command group for provisioner commands."""

    def __init__(self, environment: Environment):
        """Create CLI command group."""
        self._environment: Environment = environment

    def list(self, raw: bool = False):
        """List all provisioners."""
        provisioner_list = [
            fixture.plugin.instance_id
            for fixture in self._environment.fixtures().filter(
                interfaces=[METTA_PLUGIN_INTERFACE_ROLE_PROVISIONER]
            )
        ]

        if raw:
            return list
        return cli_output(provisioner_list)

    def _select_provisioner(self, instance_id: str = ""):
        """Pick a matching provisioner."""
        if instance_id:
            return self._environment.fixtures().get(
                interfaces=[METTA_PLUGIN_INTERFACE_ROLE_PROVISIONER],
                instance_id=instance_id,
            )
        # Get the highest priority provisioner
        return self._environment.fixtures().get(
            interfaces=[METTA_PLUGIN_INTERFACE_ROLE_PROVISIONER]
        )

    def info(self, provisioner: str = "", deep: bool = True):
        """Get info about a provisioner plugin."""
        fixture = self._select_provisioner(instance_id=provisioner)
        return cli_output(fixture.info(deep=deep))

    def fixtures(self, provisioner: str = "", deep: bool = False):
        """List all fixtures for this provisioner."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        if not hasattr(provisioner_plugin, "fixtures"):
            raise ValueError("This provisioner does not keep fixtures.")

        return cli_output(provisioner_plugin.fixtures.info(deep=deep))

    # 'up' is a common handler for provisioning.
    # pylint: disable=invalid-name
    def up(self, provisioner: str = ""):
        """Prepare and apply a provisioner."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.prepare()
        provisioner_plugin.apply()

    def prepare(self, provisioner: str = ""):
        """Run provisioner prepare."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.prepare()

    def apply(self, provisioner: str = ""):
        """Run provisioner apply."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.apply()

    def destroy(self, provisioner: str = ""):
        """Run provisioner destroy."""
        provisioner_plugin = self._select_provisioner(instance_id=provisioner).plugin
        provisioner_plugin.destroy()



"""

PLUGIN: Provisioner.

Provisioning plugins meet a common interface standard for managing a testing
system, where the provisioning keywords relate to creating and tearing down the
system.

"""

import logging

logger = logging.getLogger("metta.provisioner")

METTA_PLUGIN_INTERFACE_ROLE_PROVISIONER = "provisioner"
""" metta plugin interface identifier for client plugins """

METTA_PROVISIONER_CONFIG_PROVISIONERS_LABEL = "provisioners"
""" A centralized configerus load labe for multiple provisioners """
METTA_PROVISIONER_CONFIG_PROVISIONER_LABEL = "provisioner"
""" A centralized configerus load label a provisioner """
METTA_PROVISIONER_CONFIG_PROVISIONERS_KEY = "provisioners"
""" A centralized configerus key for multiple provisioners """
METTA_PROVISIONER_CONFIG_PROVISIONER_KEY = "provisioner"
""" A centralized configerus key for one provisioner """



"""

METTA Plugin definition and management.

The basis of plugin management is here.  The functionality allows any code to
register a plugin by decorating a factory function, or to use the factory to
ask for an instance of a plugin based on type an id.

Plugins are uniquely identified using a plugin_id string.  This string is
a parte of the plugin registration, and is used when asking the Factory to
create an instance of the plugin.

Plugins also have a list of string interfaces associated during registration
which are used to allow a plugin registration to arbitrarily decide what the
plugin can do. The interfaces descriptors play no functional role, and do
not relate to any python structures, but rather is used to allow plugin
consumers to look for plugins which advertise that they do certain things.

The order of operations for using plugins should be:

1. register the plugin by decorating a function that can build the plugin
  with the Factory class.  Provide a unique id for the plugin and also
  provide any describing string interfaces.

2. optionally use some Factory class methods to discover plugin ids.

3. ask the Factory object to create a plugin instance. This returns
  whatever the decorated factory function returns, wrapper in a small
  struct container that also contains some metadata about the plugin.

"""
import logging
import functools
from typing import List, Dict, Callable, Any

logger = logging.getLogger("metta.plugin")

METTA_PLUGIN_CONFIG_KEY_PLUGIN = "plugin"
""" configerus .get() key for a single plugin """
METTA_PLUGIN_CONFIG_KEY_PLUGINID = "plugin_id"
""" configerus .get() key for plugin_id """
METTA_PLUGIN_CONFIG_KEY_INSTANCEID = "instance_id"
""" configerus .get() key for plugin_id """
METTA_PLUGIN_CONFIG_KEY_PLUGININTERFACES = "interfaces"
""" configerus .get() key for plugin interfaces """
METTA_PLUGIN_CONFIG_KEY_PLUGINLABELS = "labels"
""" configerus .get() key for plugin labels """
METTA_PLUGIN_CONFIG_KEY_ARGUMENTS = "arguments"
""" configerus .get() key for plugin arguments """


# object is a struct (better than a Dict)
# pylint: disable=too-few-public-methods, too-many-arguments
class Instance:
    """An instances of a plugin, with some metadata about it."""

    def __init__(
        self,
        plugin_id: str,
        instance_id: str,
        interfaces: List[str],
        labels: Dict[str, str],
        plugin: object,
    ):
        """Create an Instance object."""
        self.plugin_id = plugin_id
        self.instance_id = instance_id
        self.interfaces = interfaces
        self.labels = labels
        self.plugin = plugin

    def __repr__(self) -> str:
        """Create a string representation of the instance."""
        return f"Instance({self.plugin_id}, {self.instance_id}, {self.interfaces}, {self.labels})"


# object is a struct (better than a Dict)
# pylint: disable=too-few-public-methods
class PluginInstanceFactory:
    """Struct for tracking registered plugin factories with metadata."""

    def __init__(
        self,
        plugin_id: str,
        interfaces: List[str],
        labels: Dict[str, str],
        factory_method: Callable,
    ):
        """Create a new PluginFactoryStruct."""
        self.plugin_id = plugin_id
        self.interfaces = interfaces
        self.labels = labels
        self.factory_method = factory_method


class Factory:
    """Python decorator class for metta Plugin factories.

    This class should be used to decorate any function which is meant to be a
    factory for metta plugins.

    If you are writing a plugin factory, decorate it with this class, provide
    the factory type and id values, and then the factory will be avaialble to
    other code.

    If you are trying to get an instance of a plugin, then create an instance
    of this class and use the create() method

    """

    _registry: Dict[str, PluginInstanceFactory] = {}
    """ A dict of registered factory functions."""

    def __init__(
        self, plugin_id: str, interfaces: List[str] = None, labels: Dict[str, str] = None
    ):
        """Create a new Factory instance.

        This is used in two scenarios:

        1. Execution of a decoration on a factory function.
        2. Initialization of an object that will be used to for new instances.

        Parameters:
        -----------
        plugin_id (str) : unique identifier for the plugin which this factory
            will create.  This will be used on registration and the matching
            value is used on construction.

        interfaces (List[str]) : List of string identifiers for interfaces

        Raises:
        -------
        Plugin IDs must be universally unique.  Attempting to register a
        plugin with an ID that exists already will raise a KeyError
        """
        if plugin_id in self._registry:
            raise KeyError(
                f"Plugin registration failed; there is already a plugin with the name {plugin_id}"
            )

        self._plugin_id: str = plugin_id
        self._interfaces: List[str] = interfaces if interfaces is not None else []
        self._labels: Dict[str, str] = labels if labels is not None else {}

        logger.debug("Metta Plugin factory registering `%s`", self)

    def __repr__(self) -> str:
        """Reproduce string for this object."""
        return f"Factory({self._plugin_id}, {self._interfaces}, {self._labels})"

    def __call__(self, func: Callable) -> Callable:
        """Wrap the decorator factory wrapping function.

        Returns:
        --------
        Decorated construction function.

        """
        functools.update_wrapper(self, func)

        def wrapper(*args, **kwargs):
            logger.debug("plugin factory exec: %s", self._plugin_id)
            return func(*args, **kwargs)

        self._registry[self._plugin_id] = PluginInstanceFactory(
            self._plugin_id, self._interfaces, self._labels, wrapper
        )
        return wrapper

    @classmethod
    def create(
        cls,
        plugin_id: str,
        instance_id: str,
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ) -> Any:
        """Get an instance of a plugin as created by the decorated factory.

        Parameters:
        -----------
        plugin_id (str) : plugin id of the plugin to be created, which must
            match a registered plugin id.

        akwargs : Arguments passed to the plugin factory as kwargs.

        @NOTE The metta environment object is the most common requestor of
            plugins, and it will always pass some arguments to every plugin
            factory, which means that plugins are expected to accept those
            arguments.
            When writing a plugin, know what arguments the factory is expected
            to handle.

        Returns:
        --------
        An instance of the plugin as created by the registered plugin factory.

        Metta core often expects this return to be an object instance of a
        plugin class, which will implemented interfaces based on the kind of
        plugin.  This is kind of abstract, but that follows the goals of the
        plugin management system, which is just hear to register and create
        plugins for coordinated functionality injection.

        """
        if not (isinstance(plugin_id, str) and isinstance(instance_id, str)):
            raise ValueError(
                f"Bad arguments passed for creating a fixture: " f":{plugin_id}:{instance_id}"
            )
        try:
            factory = cls._registry[plugin_id]
        except KeyError as err:
            raise NotImplementedError(
                f"METTA Plugin factory '{plugin_id}' has not been registered."
            ) from err
        except Exception as err:
            raise Exception(
                f"Could not create Plugin instance '{plugin_id}' due to an unknown error."
            ) from err

        plugin = factory.factory_method(*args, **kwargs)
        return Instance(
            plugin_id=plugin_id,
            instance_id=instance_id,
            interfaces=factory.interfaces,
            labels=factory.labels,
            plugin=plugin,
        )

    @classmethod
    def interfaces(cls) -> List[str]:
        """Return an iterable of types that have been registered.

        Returns:
        --------
        A list of string registered plugin ids

        """
        plugin_interfaces: List[str] = []
        for factory in cls._registry.values():
            for factory_interface in factory.interfaces:
                if factory_interface not in plugin_interfaces:
                    plugin_interfaces.append(factory_interface)
        return plugin_interfaces

    @classmethod
    def plugin_ids(cls, interfaces_filter: List[str] = None) -> List[str]:
        """Build a generator of matching registered plugins.

        Parameters:
        -----------
        interfaces_filter (List[str]) : if not empty, then only plugins which
            have all of the passed interfaces are included in the return.

        Returns:
        --------
        A List of plugin ids for all matching plugins (which is all plugins
        if you passed an empty filter list)
        """
        if interfaces_filter is None:
            interfaces_filter = []

        if len(interfaces_filter) == 0:
            return list(cls._registry.keys())

        plugin_ids: List[str] = []
        for factory in cls._registry.values():
            for interface_filter in interfaces_filter:
                if interface_filter not in factory.interfaces:
                    break
            else:
                # interfaces do intersect
                plugin_ids.append(factory.plugin_id)

        return plugin_ids



"""

Metta Sonobuoy package.

Primarily we register the metta plugins

"""
from typing import Any, List

from mirantis.testing.metta.plugin import Factory
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.client import METTA_PLUGIN_INTERFACE_ROLE_CLIENT
from mirantis.testing.metta_health.healthcheck import METTA_PLUGIN_INTERFACE_ROLE_HEALTHCHECK
from mirantis.testing.metta.workload import METTA_PLUGIN_INTERFACE_ROLE_WORKLOAD
from mirantis.testing.metta_cli.base import METTA_PLUGIN_INTERFACE_ROLE_CLI

from mirantis.testing.metta_kubernetes.kubeapi_client import KubernetesApiClientPlugin

from .workload import (
    SonobuoyWorkloadPlugin,
    METTA_SONOBUOY_WORKLOAD_PLUGIN_ID,
    SONOBUOY_WORKLOAD_CONFIG_LABEL,
    SONOBUOY_WORKLOAD_CONFIG_BASE,
)
from .plugin import Plugin
from .client import SonobuoyClientPlugin, METTA_SONOBUOY_CLIENT_PLUGIN_ID
from .sonobuoy import SONOBUOY_DEFAULT_RESULTS_PATH
from .cli import SonobuoyCliPlugin, METTA_PLUGIN_ID_SONOBUOY_CLI


@Factory(
    plugin_id=METTA_SONOBUOY_WORKLOAD_PLUGIN_ID,
    interfaces=[
        METTA_PLUGIN_INTERFACE_ROLE_WORKLOAD,
        METTA_SONOBUOY_WORKLOAD_PLUGIN_ID,
    ],
)
def metta_plugin_factory_workload_sonobuoy(
    environment: Environment,
    instance_id: str = "",
    label: str = SONOBUOY_WORKLOAD_CONFIG_LABEL,
    base: Any = SONOBUOY_WORKLOAD_CONFIG_BASE,
):
    """Create a metta sonobuoy workload plugin."""
    return SonobuoyWorkloadPlugin(environment, instance_id, label=label, base=base)


@Factory(
    plugin_id=METTA_SONOBUOY_CLIENT_PLUGIN_ID,
    interfaces=[
        METTA_PLUGIN_INTERFACE_ROLE_CLIENT,
        METTA_SONOBUOY_CLIENT_PLUGIN_ID,
        METTA_PLUGIN_INTERFACE_ROLE_HEALTHCHECK,
    ],
)
# pylint: disable=too-many-arguments
def metta_plugin_factory_client_sonobuoy(
    environment: Environment,
    instance_id: str,
    kubeclient: KubernetesApiClientPlugin,
    plugins: List[Plugin] = None,
    results_path: str = SONOBUOY_DEFAULT_RESULTS_PATH,
) -> SonobuoyClientPlugin:
    """Create a metta client plugin."""
    return SonobuoyClientPlugin(
        environment,
        instance_id,
        kubeclient=kubeclient,
        plugins=plugins,
        results_path=results_path,
    )


@Factory(
    plugin_id=METTA_PLUGIN_ID_SONOBUOY_CLI,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_CLI],
)
def metta_plugin_factory_cli_sonobuoy(environment: Environment, instance_id: str = ""):
    """Create a sonobuoy cli plugin."""
    return SonobuoyCliPlugin(environment, instance_id)


# ----- SetupTools EntryPoint METTA BootStrapping -----


# pylint: disable=unused-argument
def bootstrap_environment(environment: Environment):
    """METTA_Terraform bootstrap.

    Currently we only use this to import plugins.

    Parameters:
    -----------
    env (Environment) : an environment which should have validation config added to.

    """



"""

Metta CLI : Environment commands.

Various commands that allow introspection of the available environments.

"""
import logging

from mirantis.testing.metta.globals import global_fixtures
from mirantis.testing.metta.environment import Environment, METTA_PLUGIN_INTERFACE_ROLE_ENVIRONMENT

from .base import CliBase, cli_output

logger = logging.getLogger("metta.cli.environment")


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class EnvironmentCliPlugin(CliBase):
    """Fire command/group generator for environment commands."""

    def fire(self):
        """Return a dict of commands."""
        return {"environment": EnvironmentGroup(self._environment)}


class EnvironmentGroup:
    """Base Fire command group for environment commands."""

    def __init__(self, environment: Environment):
        """Create CLI command group."""
        self._environment: Environment = environment

    # needs to be a method for registration in fire
    # pylint: disable=no-self-use
    def list(self, raw: bool = False):
        """List all of the environment names."""
        names = [
            fixture.instance_id
            for fixture in global_fixtures.filter(
                interfaces=[METTA_PLUGIN_INTERFACE_ROLE_ENVIRONMENT]
            )
        ]
        if raw:
            return names
        return cli_output(names)

    def _get_environment(self, environment: str = ""):
        """Select an environment."""
        if not environment:
            environment = self._environment.instance_id()
        return global_fixtures.get(
            instance_id=environment, interfaces=[METTA_PLUGIN_INTERFACE_ROLE_ENVIRONMENT]
        )

    def name(self, environment: str = ""):
        """Return env name."""
        environment_plugin = self._get_environment(environment).plugin
        return environment_plugin.instance_id()

    def info(self, environment: str = "", deep: bool = False):
        """Return info about an environment."""
        environment_fixture = self._get_environment(environment)

        return cli_output(environment_fixture.info(deep=deep))

    def bootstraps(self, environment: str = ""):
        """List bootstraps that have been applied to the environment."""
        environment_plugin = self._get_environment(environment).plugin
        # pylint: disable=protected-access
        bootstrap_list = list(bootstrap for bootstrap in environment_plugin._environment_boostraps)

        return cli_output(bootstrap_list)



"""

Sonobuoy plugin handling.

Struct abstraction for sonobuoy plugins.

"""
from typing import Dict, List, Any
import os.path

import yaml


class Plugin:
    """Plugin definition struct.

    This makes plugin abstraction easier for parametrizing the client.

    """

    def __init__(self, plugin_id: str, plugin_def: str = "", envs: Dict[str, str] = None):
        """Parametrize the plugin definition.

        Parameters:
        -----------
        id (str): what does sonobuoy think the plugin id will be

        def (str): plugin name or path to use with the -p flag

        envs (Dict[str, str]): Dict of plugin env variables set at run.
        """
        self.plugin_id: str = plugin_id
        self.plugin_def: str = plugin_def
        self.envs: Dict[str, str] = envs if envs is not None else {}

    def run_args(self) -> List[str]:
        """Provide cli run args for the plugin.

        Returns:
        --------
        A List of str args that can be used with the sonobuoy run command.
        """
        args: List[str] = []
        args += ["--plugin", self.plugin_def]

        for key in self.envs.keys():
            args += [f"--plugin-env={self.plugin_id}.{key}={self.envs[key]}"]

        return args

    # the deep argument is a standard for the info hook
    # pylint: disable=unused-argument
    def info(self, deep: bool = True) -> Dict[str, Any]:
        """Return a Dict of info about the plugin."""
        info = {"id": self.plugin_id, "def": self.plugin_def, "env": self.envs}

        if os.path.isfile(self.plugin_def):
            info["path"] = self.plugin_def
            with open(self.plugin_def, encoding="utf-8") as def_file:
                info["def"] = yaml.safe_load(def_file)

        return info



"""

Terraform METTA provisioner plugin.

Provisioner plugin to provisioner cluster infrastructure using Terraform.

"""

import logging
import os
from typing import Any

from configerus.loaded import Loaded, LOADED_KEY_ROOT
from configerus.contrib.jsonschema.validate import (
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL,
)
from configerus.validator import ValidationError

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import (
    Fixtures,
)

from .client import TerraformClientPlugin, METTA_TERRAFORM_CLIENT_PLUGIN_ID

logger = logging.getLogger("metta_terraform:provisioner")

METTA_TERRAFORM_PROVISIONER_PLUGIN_ID = "metta_terraform_provisioner"
""" Terraform provisioner plugin id """

TERRAFORM_PROVISIONER_CONFIG_LABEL = "terraform"
""" config label loading the terraform config """
TERRAFORM_PROVISIONER_CONFIG_CHART_PATH_KEY = "plan.path"
""" config key for the terraform plan path """
TERRAFORM_PROVISIONER_CONFIG_STATE_PATH_KEY = "state.path"
""" config key for the terraform state path """
TERRAFORM_PROVISIONER_CONFIG_TFVARS_KEY = "vars"
""" config key for the terraform vars Dict, which will be written to a file """
TERRAFORM_PROVISIONER_CONFIG_TFVARS_PATH_KEY = "vars_path"
""" config key for the terraform vars file path, where the plugin will write to """
TERRAFORM_PROVISIONER_DEFAULT_TFVARS_FILE = "terraform.tfvars.json"
""" Default vars file if none was specified """
TERRAFORM_PROVISIONER_DEFAULT_STATE_SUBPATH = "metta-state"
""" Default vars file if none was specified """

TERRAFORM_VALIDATE_JSONSCHEMA = {
    "type": "object",
    "properties": {
        "type": {"type": "string"},
        "plugin_id": {"type": "string"},
        "plan": {
            "type": "object",
            "properties": {"path": {"type": "string"}},
            "required": ["path"],
        },
        "state": {"type": "object", "properties": {"path": {"type": "string"}}},
        "tfvars_path": {"type": "string"},
        "tfvars": {"type": "object"},
    },
}
""" Validation jsonschema for terraform config contents """
TERRAFORM_VALIDATE_TARGET = {
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL: TERRAFORM_VALIDATE_JSONSCHEMA
}
""" configerus validation target to match the above config """


# pylint: disable=too-many-instance-attributes
class TerraformProvisionerPlugin:
    """Terraform provisioner plugin.

    Provisioner plugin that allows control of and interaction with a terraform
    cluster.

    ## Requirements

    1. this plugin uses subprocess to call a terraform binary, so you have to
        install terraform in the environment

    ## Usage

    ### Plan

    The plan must exists somewhere on disk, and be accessible.

    You must specify the path and related configuration in config, which are
    read in the .prepare() execution.

    ### Vars/State

    This plugin reads TF vars from config and writes them to a vars file.  We
    could run without relying on vars file, but having a vars file allows cli
    interaction with the cluster if this plugin messes up.

    You can override where Terraform vars/state files are written to allow
    sharing of a plan across test suites.

    Parameters:
    -----------
    environment (Environment) : All metta plugins receive the environment
        object in which they were created.
    instance_id (str) : all metta plugins receive their own string identity.

    label (str) : Configerus load label for finding plugin config.
    base (str) : Configerus get base key in which the plugin should look for
        all config.

    """

    def __init__(
        self,
        environment: Environment,
        instance_id: str,
        label: str = TERRAFORM_PROVISIONER_CONFIG_LABEL,
        base: Any = LOADED_KEY_ROOT,
    ):
        """Run the super constructor but also set class properties.

        Interpret provided config and configure the object with all of the
        needed pieces for executing terraform commands

        """
        self._environment: Environment = environment
        """ Environemnt in which this plugin exists """
        self._instance_id: str = instance_id
        """ Unique id for this plugin instance """

        self._config_label = label
        """ configerus load label that should contain all of the config """
        self._config_base = base
        """ configerus get key that should contain all tf config """

        self.fixtures: Fixtures = Fixtures()
        """Children fixtures, typically just the client plugin."""

        # Make the client fixture in the constructor.  The TF client fixture is
        # quite state safe, and should only need to be created once, unlike
        # other provisioner clients which may be vulnerable to state change.
        self.make_fixtures()

    # deep argument is an info() standard across plugins
    # pylint: disable=unused-argument
    def info(self, deep: bool = False):
        """Get info about the provisioner plugin.

        Returns:
        --------
        Dict of keyed introspective information about the plugin.

        """
        terraform_config: Loaded = self._environment.config().load(self._config_label)

        info = {
            "config": {
                "label": self._config_label,
                "base": self._config_base,
                "tfvars": terraform_config.get(
                    [self._config_base, TERRAFORM_PROVISIONER_CONFIG_TFVARS_KEY],
                    default="NONE",
                ),
                "chart_path": terraform_config.get(
                    [self._config_base, TERRAFORM_PROVISIONER_CONFIG_CHART_PATH_KEY],
                    default="MISSING",
                ),
                "state_path": terraform_config.get(
                    [self._config_base, TERRAFORM_PROVISIONER_CONFIG_STATE_PATH_KEY],
                    default="MISSING",
                ),
                "tfvars_path": terraform_config.get(
                    [self._config_base, TERRAFORM_PROVISIONER_CONFIG_TFVARS_PATH_KEY],
                    default="MISSING",
                ),
            },
            "client": {
                "instance_id": self.client_instance_id(),
            },
        }

        return info

    def prepare(self):
        """Run terraform init."""
        logger.info("Running Terraform INIT")
        self._get_client_plugin().init()

    def apply(self, lock: bool = True):
        """Create all terraform resources described in the plan."""
        logger.info("Running Terraform APPLY")
        self._get_client_plugin().apply(lock=lock)

    def destroy(self, lock: bool = True):
        """Remove all terraform resources in state."""
        logger.info("Running Terraform DESTROY")
        self._get_client_plugin().destroy(lock=lock)
        # accessing parent property for clearing out existing output fixtures
        # pylint: disable=attribute-defined-outside-init
        self.fixtures = Fixtures()

    def make_fixtures(self):
        """Make the client plugin for terraform interaction."""
        try:
            terraform_config = self._environment.config().load(
                self._config_label, force_reload=True, validator=TERRAFORM_VALIDATE_TARGET
            )
            """ get a configerus LoadedConfig for the label """
        except ValidationError as err:
            raise ValueError("Terraform config failed validation") from err

        try:
            chart_path = terraform_config.get(
                [self._config_base, TERRAFORM_PROVISIONER_CONFIG_CHART_PATH_KEY]
            )
            """ subprocess commands for terraform will be run in this path """
        except Exception as err:
            raise ValueError(
                "Plugin config did not give us a working/plan path:" f" {terraform_config.get()}"
            ) from err

        state_path = terraform_config.get(
            [self._config_base, TERRAFORM_PROVISIONER_CONFIG_STATE_PATH_KEY],
            default=os.path.join(chart_path, TERRAFORM_PROVISIONER_DEFAULT_STATE_SUBPATH),
        )
        """ terraform state path """

        tfvars = terraform_config.get(
            [self._config_base, TERRAFORM_PROVISIONER_CONFIG_TFVARS_KEY],
            default={},
        )
        """ List of vars to pass to terraform.  Will be written to a file """

        tfvars_path = terraform_config.get(
            [self._config_base, TERRAFORM_PROVISIONER_CONFIG_TFVARS_PATH_KEY],
            default=os.path.join(chart_path, TERRAFORM_PROVISIONER_DEFAULT_TFVARS_FILE),
        )
        """ vars file which will be written before running terraform """

        logger.debug("Creating Terraform client")

        fixture = self._environment.new_fixture(
            plugin_id=METTA_TERRAFORM_CLIENT_PLUGIN_ID,
            instance_id=self.client_instance_id(),
            priority=70,
            arguments={
                "chart_path": chart_path,
                "state_path": state_path,
                "tfvars": tfvars,
                "tfvars_path": tfvars_path,
            },
            labels={
                "parent_plugin_id": METTA_TERRAFORM_PROVISIONER_PLUGIN_ID,
                "parent_instance_id": self._instance_id,
            },
            replace_existing=True,
        )
        # keep this fixture attached to the workload to make it retrievable.
        self.fixtures.add(fixture, replace_existing=True)

    def client_instance_id(self) -> str:
        """Construct an instanceid for the child client plugin."""
        return f"{self._instance_id}-{METTA_TERRAFORM_CLIENT_PLUGIN_ID}"

    def _get_client_plugin(self) -> TerraformClientPlugin:
        """Retrieve the client plugin if we can."""
        return self.fixtures.get_plugin(plugin_id=METTA_TERRAFORM_CLIENT_PLUGIN_ID)



"""

PLUGIN: Output.

Workload plugins use clients to apply workloads to a testing system.  The
system is irrelevant tothe workload plugin, as it only cares whether or not it
can retrieve the expected clients.

Workload clients are expected to allow creation of Workload instances from
clients, so that multiple applications of the workload can be managed at one
time.

"""

METTA_PLUGIN_INTERFACE_ROLE_WORKLOAD = "workload"
""" Metta plugin interface identifier for workloads """

METTA_WORKLOAD_CONFIG_WORKLOADS_LABEL = "workloads"
""" A centralized configerus load label for multiple workloads """
METTA_WORKLOAD_CONFIG_WORKLOAD_LABEL = "workload"
""" A centralized configerus load label for a workload """
METTA_WORKLOAD_CONFIG_WORKLOADS_KEY = "workloads"
""" A centralized configerus key for multiple workloads """
METTA_WORKLOAD_CONFIG_WORKLOAD_KEY = "workload"
""" A centralized configerus key for one workload """



"""

SetupTools implementation for all of the mirantis Metta packages

does not include the individual suites, which each act as their own packages.

"""
from setuptools import setup

if __name__ == "__main__":
    setup()



"""

Bootstrapping Metta when it starts to run.

Separate abstraction for boostrapping python modules by executing setuptools
entrypoints with arbitrary arguments.

"""
import logging
from typing import List, Dict, Any
import os

from configerus.config import Config

from .plugin import Factory
from .fixture import Fixture
from .globals import global_fixtures
from .building import FixtureBuildingFromConfigMixin, FixtureBuildingFromDictMixin
from .discover import discover_project_root
from .config import add_config_sources_from_config, METTA_CONFIG_CONFIG_SOURCE_KEY
from .importing import add_imports_from_config, METTA_IMPORT_CONFIG_LABEL
from .setuptools import setuptools_entrypoint, METTA_CONFIG_SETUPTOOLS_BOOTSTRAPS_KEY
from .environment import METTA_FIXTURES_CONFIG_ENVIRONMENTS_KEY

logger = logging.getLogger("metta.bootstrapper")

METTA_ENTRYPOINT_BOOTSTRAPPER = "metta.bootstrap.bootstrapper"
""" SetupTools entry_point used for METTA bootstrapping bootstrappers """

METTA_PLUGIN_INTERFACE_ROLE_BOOTSTRAPPER = "bootstrapper"
""" metta plugin interface identifier for environment plugins """

METTA_BOOTSTRAPPER_CORE_PLUGIN_ID = "core_bootstrap"
"""A core default bootstrapper."""
METTA_BOOTSTRAPPER_PROJECT_PLUGIN_ID = "project_bootstrap"
"""A core default bootstrapper."""

METTA_BOOTSTRAP_PROJECT_CONFIG_LABEL = "metta"
"""Configerus .load() label for finding project bootstrapping config."""

CWD = os.path.realpath(os.getcwd())
""" CWD if needed for project discovery """


@Factory(
    plugin_id=METTA_BOOTSTRAPPER_CORE_PLUGIN_ID,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_BOOTSTRAPPER],
)
def core_bootstrap(config: Config, instance_id: str) -> "BootStrapFixtureBuilder":
    """Core default bootstrapper.

    This bootstrapper works by assuming that enough config is in place for it to
    be able to create environments, and that those environments know how to
    create any fixtures excepted based on their config.
    This sounds elegant, but of course it means that the user needs to know how
    to write config in the right way. At least the config is very standardized
    and repeated, so all you really need to know is:

    1. there should be config under the label "metta".
    2. that config can optionally direct us to include more config
    3. that config can optionally direct us to import some python (which is good
        for importing modules that contain plugins.)
    4. that config should probable container an "environments" base which is an
        array of config objects that define environment plugin instances.
    5. Those environments can also repeat #2 and #3 above,
    6 Those environments should probably contain "fixture" bases as arrays of
        plugins that we should create.

    """
    # 1. add any config sources specified in config
    # (Here config tells us to load more config. It is weird but ok.)
    add_config_sources_from_config(
        config=config,
        label=METTA_BOOTSTRAP_PROJECT_CONFIG_LABEL,
        base=METTA_CONFIG_CONFIG_SOURCE_KEY,
    )

    # 2. import any python code requested, to make sure that all funxtionality
    #    is in scope that is needed.
    #    Here we say look in the "metta" config for a "imports" section.
    add_imports_from_config(
        config=config, label=METTA_BOOTSTRAP_PROJECT_CONFIG_LABEL, base=METTA_IMPORT_CONFIG_LABEL
    )

    # create a bootstrap builder which will build plugins in global scope
    builder: BootStrapFixtureBuilder = BootStrapFixtureBuilder(
        config=config, instance_id=instance_id
    )

    # 3. run any bootstrapper setuptools bootstraps
    bootstrapper_bootstraps: List[str] = config.load(METTA_BOOTSTRAP_PROJECT_CONFIG_LABEL).get(
        METTA_CONFIG_SETUPTOOLS_BOOTSTRAPS_KEY, default=[]
    )
    setuptools_entrypoint(
        entrypoint=METTA_ENTRYPOINT_BOOTSTRAPPER,
        entries=bootstrapper_bootstraps,
        args=[builder],
        kwargs={},
    )

    # 4. build environments from config
    #    Environment plugins are top level containers, and should build their
    #    own plugins on creation.
    try:
        labels: Dict[str, str] = {
            "container": "bootstrapper",
            "container_id": instance_id,
            "bootstrapper": instance_id,
        }
        builder.add_fixtures_from_config(
            label=METTA_BOOTSTRAP_PROJECT_CONFIG_LABEL,
            base=METTA_FIXTURES_CONFIG_ENVIRONMENTS_KEY,
            labels=labels,
            exception_if_missing=True,
        )
    except KeyError as err:
        raise RuntimeError(
            f"Bootstrapper encountered an issue creating environments from config: {err}"
        ) from err

    return builder


@Factory(
    plugin_id=METTA_BOOTSTRAPPER_PROJECT_PLUGIN_ID,
    interfaces=[METTA_PLUGIN_INTERFACE_ROLE_BOOTSTRAPPER],
)
def project_bootstrap(
    config: Config, instance_id: str, path: str = CWD
) -> "BootStrapFixtureBuilder":
    """Bootstrap Metta by looking in a file-system for a project.

    First we look in the file system for some config files, and use that
    file path and config as a source of direction on how to bootstrap Metta.

    That file source can be like any fixture building config in that it can
    instructions for importing python modules, describing config sources,
    and building fixtures, but any plugins that required an environment may
    fail as the fixture builder does not pass in an Environment.

    Parameters:
    -----------
    config (Config) : Configerus Config object used internally and by any
        functionality that wants to consider itself inside the environment.

    """
    # 1. Discover a root project config path, which gives us enough config to start
    discover_project_root(config, start_path=path)

    # At this point, the setup is done as we should have found some standardized
    # config which instructs the core bootloader how to do its magic.
    # @see core_bootstrap above

    # 2. run the core bootstrapper
    instance_id = f"{instance_id}-core"
    args: List[Any] = [config, instance_id]
    kwargs: Dict[str, Any] = {}
    return Factory.create(METTA_BOOTSTRAPPER_CORE_PLUGIN_ID, instance_id, *args, **kwargs)


class BootStrapFixtureBuilder(FixtureBuildingFromConfigMixin, FixtureBuildingFromDictMixin):
    """A utility tool to build bootstrap fixtures in global scope."""

    def __init__(self, config: Config, instance_id: str):
        """Get a config object.

        Parameters:
        -----------
        config (Config) : Configerus Config object used internally and by any
            functionality that wants to consider itself inside the environment.

        """
        self._config = config
        """Keep the config object in scope for fixture building."""
        self._instance_id: str = instance_id
        """Use this builder name for reporting."""

        FixtureBuildingFromConfigMixin.__init__(
            self, config=config, builder_callback=self.new_fixture
        )
        FixtureBuildingFromDictMixin.__init__(self, builder_callback=self.new_fixture)

    # This is what it takes to build a plugin
    # pylint: disable=too-many-arguments
    def new_fixture(
        self,
        plugin_id: str,
        instance_id: str,
        priority: int,
        arguments: Dict[str, Any] = None,
        labels: Dict[str, Any] = None,
        replace_existing=False,
    ) -> Fixture:
        """Create a new global plugin from parameters.

        Parameters:
        -----------
        plugin_id (str) : METTA plugin id to tell us what plugin factory to use;

            @see .plugin.Factory for more details on how plugins are loaded.

        instance_id (str) : string instance id that will be passed to the new
            plugin object;

        priority (int) : Integer priority 1-100 for comparative prioritization
            between other plugins;

        arguments (Dict[str, Any]) : Keyword Arguments which should be passed to
            the plugin constructor after environment and instance_id;

        labels (Dict[str, str]) : Keyword/value labels which are to be associated
            with the plugin/fixture;

        replace_existing (bool) : Replace any existing matching fixture.

        Return:
        -------
        A Fixture object with the new plugin added

        The Fixtures has already been added to the global fixtures, but is returned
        so that the consumer can act on it separately without haveing to
        search for it.

        Raises:
        -------
        NotImplementedError if you asked for an unregistered plugin_id

        """
        if arguments is None:
            arguments = {}

        # Catch some early arg validation errors which would otherwise have
        # caused some hard to diagnose issues.
        if not (
            isinstance(plugin_id, str)
            and isinstance(instance_id, str)
            and isinstance(priority, int)
        ):
            raise ValueError(
                f"Bad arguments passed for creating a fixture: "
                f":{plugin_id}:{instance_id} ({priority})"
            )

        if labels is None:
            labels = {}
        labels["bootstrapper"] = self._instance_id

        # Build the plugin instance by passing collected arguments to the
        # registered plugin factory. This will call whatever function was
        # decorated for the plugin_id.
        args: List[Any] = [self._config, instance_id]
        kwargs: Dict[str, Any] = arguments
        plugin_instance = Factory.create(plugin_id, instance_id, *args, **kwargs)

        # Build a fixture from the plugin_instance and add it to the fixtures
        # set for the environment, then return the fixture.
        fixture = global_fixtures.add(
            fixture=Fixture.from_instance(plugin_instance, priority=priority, labels=labels),
            replace_existing=replace_existing,
        )
        return fixture



"""

Run a Sonobuoy run on a k82 client.

Use this to run the sonobuoy implementation

"""
from typing import Dict, Any, List
import logging
import subprocess
import os
import shutil
import tempfile

import kubernetes

from mirantis.testing.metta_kubernetes.kubeapi_client import KubernetesApiClientPlugin

from .plugin import Plugin
from .results import SonobuoyResults, SonobuoyStatus

logger = logging.getLogger("sonobuoy")

SONODBUOY_DEFAULT_WAIT_PERIOD_SECS = 1440
""" Default time for sonobuoy to wait when running """
SONODBUOY_DEFAULT_WAIT_OUTPUT = "progress"
"""What type of --wait-ouput to pass to sonobuoy run with --wait."""
SONOBUOY_DEFAULT_BIN = "sonobuoy"
""" Default Bin Name for running sonobuoy """
SONOBUOY_DEFAULT_RESULTS_PATH = "./results"
""" Default path for where to download sonobuoy results """
SONOBUOY_NAMESPACE = "sonobuoy"
"""K8s namespace where sonobuoy puts its stuff (RO)."""

SONOBUOY_CRB_NAME = "sonobuoy-serviceaccount-cluster-admin"
"""Sonobuoy cluster-role-binding name."""
SONOBUOY_CRB_ROLEREF_KIND = "ClusterRole"
"""Sonobuoy cluster-role-binding kind."""
SONOBUOY_CRB_ROLEREF_NAME = "cluster-admin"
"""Sonobuoy cluster-role-binding role."""
SONOBUOY_CRB_SUBJECTS_SERVICEACCOUNT = "sonobuoy-serviceaccount"
"""Sonobuoy cluster-role-binding serviceaccount name."""


# pylint: disable=too-many-instance-attributes
class SonobuoyClient:
    """A sonobuoy handler."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        kubeclient: KubernetesApiClientPlugin,
        plugins: List[Plugin] = None,
        binary: str = SONOBUOY_DEFAULT_BIN,
        results_path: str = SONOBUOY_DEFAULT_RESULTS_PATH,
        create_crbs: bool = True,
    ):
        """Initialize the workload instance."""
        self._api_client = kubeclient
        """Kube API metta client, used for a config file and for creating K8s resources."""
        self.kubeconfig = kubeclient.config_file
        """String path to a kubeconfig file."""

        self.plugins = plugins if plugins is not None else []
        """Which sonobuoy plugins to run."""

        if shutil.which(binary) is None:
            raise ValueError(
                "Sonobuoy binary not found. Sonobuoy commands cannot be called. "
                f"Expected binary at path {binary}"
            )

        self.bin = binary
        """Path to the sonobuoy binary."""

        self.results_path = os.path.realpath(results_path)
        """Path to where to download sonobuoy results."""

        self.create_crbs: bool = create_crbs
        """Whether or not this client should create sonobuoy CRB resources."""

    # deep is a metta info standard and expected to be here
    # pylint: disable=unused-argument
    def info(self, deep: bool = False) -> Dict[str, Any]:
        """Provide a Dict of info about the client."""
        return {
            "config": {
                "kubeconfig": self.kubeconfig,
                "results_path": self.results_path,
            },
            "plugins": [plugin.info(deep=deep) for plugin in self.plugins],
            "client": {
                "sonobuoy_bin_path": self.bin,
            },
        }

    def run(self, wait: bool = True, run_args: List[str] = None):
        """Run sonobuoy."""
        cmd = ["run"]

        if run_args is not None:
            cmd += run_args

        if self.plugins:
            for plugin in self.plugins:
                cmd += plugin.run_args()

        if wait:
            cmd += [f"--wait={SONODBUOY_DEFAULT_WAIT_PERIOD_SECS}"]
            cmd += [f"--wait-output={SONODBUOY_DEFAULT_WAIT_OUTPUT}"]

        try:
            if self.create_crbs:
                logger.info("Ensuring that we have needed K8s CRBs")
                self.create_k8s_crb()
            logger.info("Starting Sonobuoy run : %s", cmd)
            self._run(cmd)
        except subprocess.CalledProcessError as err:
            raise RuntimeError("Sonobuoy RUN failed") from err

    def status(self) -> "SonobuoyStatus":
        """Retrieve Sonobuoy status return."""
        cmd = ["status", "--json"]
        status = self._run(cmd, return_output=True)
        if status is None:
            raise ValueError("Sonobuoy did not return a status.")
        return SonobuoyStatus(status)

    def logs(self, follow: bool = True):
        """Output sonobuoy logs."""
        cmd = ["logs"]

        if follow:
            cmd += ["--follow"]

        self._run(cmd)

    def retrieve(self) -> "SonobuoyResults":
        """Retrieve sonobuoy results."""
        logger.debug("retrieving sonobuoy results to %s", self.results_path)
        try:
            os.makedirs(self.results_path, exist_ok=True)
            cmd = ["retrieve", self.results_path]
            file = self._run(cmd=cmd, return_output=True).rstrip("\n")
            if not os.path.isfile(file):
                raise RuntimeError("Sonobuoy did not retrieve a results tarball.")
            return SonobuoyResults(tarball=file, folder=self.results_path)
        except Exception as err:
            raise RuntimeError("Could not retrieve sonobuoy results") from err

    def results(self) -> str:
        """Report on sonobuoy results."""
        logger.debug("retrieving sonobuoy results to %s", self.results_path)
        try:
            with tempfile.TemporaryDirectory() as tmpdirname:
                print("created temporary directory", tmpdirname)
                cmd = ["retrieve", tmpdirname]
                file_name = self._run(cmd=cmd, return_output=True).rstrip("\n")
                if not os.path.isfile(file_name):
                    raise RuntimeError("Sonobuoy did not retrieve a results tarball.")
                cmd = ["results", os.path.join(tmpdirname, file_name)]
                return self._run(cmd=cmd, include_kubeconfig=False, return_output=True)
        except Exception as err:
            raise RuntimeError("Could not retrieve sonobuoy results") from err

    # pylint: disable=redefined-builtin
    def delete(self, all: bool = True, wait: bool = False):
        """Delete sonobuoy resources."""
        cmd = ["delete"]

        if wait:
            cmd += ["--wait"]
            cmd += [f"--wait-output={SONODBUOY_DEFAULT_WAIT_OUTPUT}"]

        self._run(cmd)
        if self.create_crbs:
            self.delete_k8s_crb()

    def version(self) -> Dict[str, str]:
        """Retrieve sonobuoy version info."""
        cmd = ["version"]
        version_string = self._run(cmd, return_output=True)
        version: Dict[str, str] = {}
        for version_item_string in version_string.strip().split("\n"):
            version_item_list = version_item_string.split(":")
            key = version_item_list[0].strip()
            value = version_item_list[1].strip()
            version[key] = value
        return version

    def _run(
        self,
        cmd: List[str],
        ignore_errors: bool = True,
        return_output: bool = False,
        include_kubeconfig: bool = True,
    ):
        """Run a sonobuoy command."""
        cmd = [self.bin] + cmd

        if include_kubeconfig:
            cmd += [f"--kubeconfig={self.kubeconfig}"]

        # this else makes it much more readable
        # pylint: disable=no-else-return
        if return_output:
            logger.debug("running sonobuoy command with output capture: %s", " ".join(cmd))
            return_res = subprocess.run(cmd, shell=False, check=True, stdout=subprocess.PIPE)

            # sonobuoy's uses of subprocess error is overly inclusive for us
            if not ignore_errors:
                return_res.check_returncode()

            return return_res.stdout.decode("utf-8")

        else:
            logger.warning("running sonobuoy command: %s", " ".join(cmd))
            res = subprocess.run(cmd, check=True, text=True)

            if not ignore_errors:
                res.check_returncode()

            return res

    def create_k8s_crb(self):
        """Create the cluster role binding that sonobuoy needs."""
        rbac_authorization_v1_api = self._api_client.get_api("RbacAuthorizationV1Api")

        # if the CRB does not exist then create it.
        try:
            return rbac_authorization_v1_api.read_cluster_role_binding(
                name=SONOBUOY_CRB_NAME
            ).to_dict()
        except kubernetes.client.exceptions.ApiException:
            logger.debug("Sonobuoy CRB not found. Creating it now.")
            body = kubernetes.client.V1ClusterRoleBinding(
                metadata=kubernetes.client.V1ObjectMeta(
                    name=SONOBUOY_CRB_NAME,
                ),
                subjects=[
                    kubernetes.client.V1Subject(
                        kind="ServiceAccount",
                        name=SONOBUOY_CRB_SUBJECTS_SERVICEACCOUNT,
                        namespace="sonobuoy",
                    )
                ],
                role_ref=kubernetes.client.V1RoleRef(
                    kind="ClusterRole",
                    name=SONOBUOY_CRB_ROLEREF_NAME,
                    api_group="rbac.authorization.k8s.io",
                ),
            )

            try:
                return rbac_authorization_v1_api.create_cluster_role_binding(body=body)
            except kubernetes.client.exceptions.ApiException as err:
                raise RuntimeError("Sonobuoy could not create the needed K8s CRB.") from err

    def delete_k8s_crb(self):
        """Remove the cluster role binding that we created."""
        rbac_authorization_v1_api = self._api_client.get_api("RbacAuthorizationV1Api")

        try:
            return rbac_authorization_v1_api.delete_cluster_role_binding(
                name=SONOBUOY_CRB_NAME
            ).to_dict()
        except kubernetes.client.exceptions.ApiException as err:
            logger.error("Could not delete sonobuoy CRB: %s", err)
            return None



"""

Metta client plugin for testkit.

"""

import logging
from typing import Any, Dict, List

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import Fixtures

from mirantis.testing.metta_mirantis.mke_client import (
    MKEAPIClientPlugin,
    METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID,
)
from mirantis.testing.metta_mirantis.msr_client import METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID

from .testkit import TestkitClient

logger = logging.getLogger("testkit.client")

METTA_TESTKIT_CLIENT_PLUGIN_ID = "metta_testkit_client"
""" Metta plugin id for testkit provisioner plugins """


class TestkitClientPlugin:
    """Testkit client plugin.

    client plugin that allows control of and interaction with a testkit
    cluster.

    ## Requirements

    1. this plugin uses subprocess to call a testkit binary, so you have to
       install testkit in the environment

    ## Usage

    @TODO

    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        environment,
        instance_id,
        system_name: str,
        config_file: str,
        systems: Dict[str, Dict[str, str]] = None,
    ):
        """Initialize Testkit provisioner.

        Parameters:
        -----------
        environment (Environment) : metta environment object that this plugin
            is attached.
        instance_id (str) : label for this plugin instances.
        config_file (str) : string path to the testkit config file.
        systems (Dict[str, Dict[str, str]]) : A dictionary of systems which this
            client is expected to provide using testkit.

            This is something which should be determinable using the config/file
            client directly, but sits outside of information encapsulated in
            the tool/conf.

            What we are talking about here is information to answer questions:

                Did testkit install MKE? if so, what accesspoint and U/P can I
                use to build an MKE client to access it.

            This is not an ideal approach but rather a necessity.

        """
        self._environment: Environment = environment
        """ Environemnt in which this plugin exists """
        self._instance_id: str = instance_id
        """ Unique id for this plugin instance """

        self._system_name: str = system_name
        """ What will testkit call the system, used client ops """

        self._testkit = TestkitClient(config_file=config_file)
        """ testkit client object """

        self._systems = systems
        """What systems will testkit install, so what fixtures are needed"""

        self.fixtures = Fixtures()
        """This object makes and keeps track of fixtures for MKE/MSR clients."""
        try:
            self.make_fixtures()
            # pylint: disable= broad-except
        except Exception as err:
            # there are many reasons this can fail, and we just want to
            # see if we can get fixtures early.
            # No need to ask forgiveness for this one.
            logger.debug("Could not make initial fixtures: %s", err)

    # the deep argument is a standard for the info hook
    # pylint: disable=unused-argument
    def info(self, deep: bool = False) -> Dict[str, Any]:
        """Get info about a provisioner plugin."""
        info = {
            "plugin": {
                "system_name": self._system_name,
            },
            "client": self._testkit.info(deep=deep),
        }
        if deep:
            try:
                info["hosts"] = self.hosts()
            # pylint: disable=broad-except
            except Exception:
                pass

        return info

    def version(self):
        """Return testkit client version."""
        return self._testkit.version()

    def create(self, opts: List[str]):
        """Run the testkit create command."""
        self._testkit.create(opts=opts)
        self.make_fixtures()

        mke_plugin = self._get_mke_client_plugin()
        mke_plugin.api_get_bundle(force=True)
        mke_plugin.make_fixtures()

    def destroy(self):
        """Remove a system from testkit."""
        return self._testkit.system_rm(system_name=self._system_name)

    def hosts(self):
        """List testkit system machines."""
        return self._testkit.machine_ls(system_name=self._system_name)

    def exec(self, host: str, cmd: str):
        """List testkit system machines."""
        return self._testkit.machine_ssh(machine=host, cmd=cmd)

    def system_ls(self):
        """List all of the systems testkit can see using our config."""
        return self._testkit.system_ls()

    # pylint: disable=too-many-branches
    def make_fixtures(self):
        """Make related fixtures from a testkit installation.

        Creates:
        --------

        MKE client : if we have manager nodes, then we create an MKE client
            which will then create docker and kubernestes clients if they are
            appropriate.

        MSR Client : if we have an MSR node, then the related client is
            created.

        """
        if self._systems is None:
            return

        testkit_hosts = self._testkit.machine_ls(system_name=self._system_name)
        """ list of all of the testkit hosts. """

        manager_hosts = []
        worker_hosts = []
        mke_hosts = []
        msr_hosts = []
        for host in testkit_hosts:
            host["address"] = host["public_ip"]
            if host["swarm_manager"] == "yes":
                manager_hosts.append(host)
            else:
                worker_hosts.append(host)

            if host["ucp_controller"] == "yes":
                mke_hosts.append(host)

        if len(msr_hosts) == 0 and len(worker_hosts) > 0:
            # Testkit installs MSR on the first work node, but the api is
            # accessible using port 444 in order to not conflict.
            first_worker = worker_hosts[0]
            first_worker_ip = first_worker["public_ip"]
            first_worker["msr_accesspoint"] = f"{first_worker_ip}:444"
            msr_hosts.append(first_worker)

        if len(mke_hosts) > 0 and METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID in self._systems:
            instance_id = f"{self._instance_id}-{METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID}"
            arguments = self._systems[METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID]
            arguments["hosts"] = mke_hosts

            if "accesspoint" in arguments and arguments["accesspoint"]:
                arguments["accesspoint"] = clean_accesspoint(arguments["accesspoint"])

            logger.debug("Launchpad client is creating an MKE client plugin: %s", instance_id)
            fixture = self._environment.new_fixture(
                plugin_id=METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID,
                instance_id=instance_id,
                priority=70,
                arguments=arguments,
                labels={
                    "parent_plugin_id": METTA_TESTKIT_CLIENT_PLUGIN_ID,
                    "parent_instance_id": self._instance_id,
                },
                replace_existing=True,
            )
            self.fixtures.add(fixture, replace_existing=True)

            # We got an MKE client, so let's activate it.

        else:
            logger.debug("No MKE master hosts found, not creating an MKE client.")

        if len(msr_hosts) > 0 and METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID in self._systems:
            instance_id = f"{self._instance_id}-{METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID}"
            arguments = self._systems[METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID]
            arguments["hosts"] = msr_hosts

            if "accesspoint" in arguments and arguments["accesspoint"]:
                arguments["accesspoint"] = clean_accesspoint(arguments["accesspoint"])

            logger.debug("Launchpad client is creating an MSR client plugin: %s", instance_id)
            fixture = self._environment.new_fixture(
                plugin_id=METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID,
                instance_id=instance_id,
                priority=70,
                arguments=arguments,
                labels={
                    "parent_plugin_id": METTA_TESTKIT_CLIENT_PLUGIN_ID,
                    "parent_instance_id": self._instance_id,
                },
                replace_existing=True,
            )
            self.fixtures.add(fixture, replace_existing=True)

        else:
            logger.debug("No MSR master hosts found, not creating an MSR client.")

    def _get_mke_client_plugin(self) -> MKEAPIClientPlugin:
        """Retrieve the MKE client plugin if we can."""
        try:
            return self.fixtures.get_plugin(plugin_id=METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID)
        except KeyError as err:
            raise RuntimeError(
                "Launchpad client cannot find its MKE client plugin, and "
                "cannot process any client actions.  Was a client created?"
            ) from err


def clean_accesspoint(accesspoint: str) -> str:
    """Remove any https:// and end / from an accesspoint."""
    accesspoint = accesspoint.replace("https://", "")
    return accesspoint



"""

METTA PLUGIN: bin download helper.

A helper plugin for downloading executables from the internet to help other
Plugins which might rely on external executables.

@NOTE remote urls may be tarballs/zips

@NOTE it goes without saying that downloading executables comes with an inherit
risk.  The request validates any certs, but it is up to you to make sure that
no config is in place which downloads dangerous bins.

The idea here is that you can tell your system to put bins from a list of
urls into a path, and have that path included in PATH in your python
environment. If you want, use a common system path like '/usr/local/bin' or
use a temporary or local path if you want to isolate versions for your test
suite.

There are two ways you can use this:

1. put a list of bins into the plugin config, and it will download when it is
constructed.  This de-couples acquiring bins from plugins that may use them, and
does all of the downloading up front before all fixtures are created.
If another fixture needs a bin in its own constructor, just make sure that this
fixture is created first.

2. any other code with access to the environment can access the binhelper
plugin and provide details to the `get_bin` method.  This couples the plugins,
and means that you will have to ensure that this plugin is available and handle
dependencies yourself.

"""
from typing import Dict
import shutil
import os
import logging
import tarfile
import zipfile
import re
import stat
import platform

import requests

from configerus.loaded import LOADED_KEY_ROOT
from configerus.contrib.jsonschema.validate import (
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL,
)
from configerus.validator import ValidationError

from mirantis.testing.metta.environment import Environment

logger = logging.getLogger("metta.common.utility:binhelper")

METTA_PLUGIN_ID_UTILITY_BINHELPER = "bin-helper-utility"
""" metta utility plugin_id for the bin-helper plugin """

BINHELPER_UTILITY_CONFIG_LABEL = "binhelper"
""" Config label used by default to load binhelper configuration """

BINHELPER_UTILITY_CONFIG_BASE_LOCALPATH = "path.local"
""" config base for what the local bin path should be """
BINHELPER_UTILITY_CONFIG_BASE_ADDTOPATH = "path.add_to_path"
""" config base for if we need to modify the env PATH and add the path """
BINHELPER_UTILITY_CONFIG_BASE_PLATFORMS = "platforms"
""" config base for list of platform bins to load on construction """

BINHELPER_UTILITY_CONFIG_BASE_BIN_URL = "url"
""" config base inside bin for bin url for downloading """
BINHELPER_UTILITY_CONFIG_BASE_BIN_VERSION = "version"
""" config base inside bin for bin version """
BINHELPER_UTILITY_CONFIG_BASE_BIN_COPYPATHS = "copy"
""" config base inside bin for a map of bins name->path-inpackage """

BINHELPER_CONFIG_JSONSCHEMA = {
    "type": "object",
    "path": {
        "type": "object",
        "properties": {"local": {"type": "string"}, "add_to_environ": {"type": "bool"}},
        "required": ["path"],
    },
    "platforms": {"$ref": "#/definitions/platform"},
    "definitions": {
        "bin": {
            "type": "object",
            "properties": {
                "url": {"type": "string"},
                "version": {"type": "string"},
                "copy": {"type": "object"},
            },
            "required": ["url"],
        },
        "platform": {"$ref": "#/definitions/bin"},
    },
    "required": ["path", "platforms"],
}
""" JSONSCHEMA for the binhelper config """
BINHELPER_CONFIG_VALIDATE_TARGET = {
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL: BINHELPER_CONFIG_JSONSCHEMA
}
""" configerus validate target for binhelper config """


# This needs to be a metta plugin
# pylint: disable=too-few-public-methods
class DownloadableExecutableUtility:
    """A helper to help you download executables."""

    def __init__(
        self,
        environment: Environment,
        instance_id: str,
        label: str = BINHELPER_UTILITY_CONFIG_LABEL,
        base: str = LOADED_KEY_ROOT,
    ):
        """Perform initial plugin configuration.

        Parameters:
        -----------
        envirionment (Environment) : environment in which this plugin exists.
        instance_id (str) : string id of this plugin instance.

        label (str) : Configerus config load label for accessing plugin
            instance config.
        base (Any) : Configerus config get key for a base that contains all
            plugin instance config.

        """
        self._environment: Environment = environment
        """ Environemnt in which this plugin exists """
        self._instance_id: str = instance_id
        """ Unique id for this plugin instance """

        loaded = self._environment.config().load(label)
        try:
            loaded.get(base, validator=BINHELPER_CONFIG_VALIDATE_TARGET)
        except KeyError as err:
            raise ValueError("Bin-Helper configuration is missing") from err
        except ValidationError as err:
            raise ValueError("Bin-Helper received invalid configuration") from err

        self.local_path = loaded.get([base, BINHELPER_UTILITY_CONFIG_BASE_LOCALPATH])

        add_to_path = loaded.get([base, BINHELPER_UTILITY_CONFIG_BASE_ADDTOPATH])
        if add_to_path:
            os.environ["PATH"] += os.pathsep + os.path.realpath(self.local_path)

        platforms = loaded.get([base, BINHELPER_UTILITY_CONFIG_BASE_PLATFORMS])
        current_platform = f"{platform.system()}-{platform.machine()}"

        if current_platform not in platforms:
            logger.warning(
                "BinHelper doesn't have configuration for your platform '%s', so it "
                "won't download any bins.",
                current_platform,
            )
            return

        bins = platforms[current_platform]
        for bin_id in list(bins.keys()):
            url = loaded.get(
                [
                    base,
                    BINHELPER_UTILITY_CONFIG_BASE_PLATFORMS,
                    current_platform,
                    bin_id,
                    BINHELPER_UTILITY_CONFIG_BASE_BIN_URL,
                ]
            )
            copypaths = loaded.get(
                [
                    base,
                    BINHELPER_UTILITY_CONFIG_BASE_PLATFORMS,
                    current_platform,
                    bin_id,
                    BINHELPER_UTILITY_CONFIG_BASE_BIN_COPYPATHS,
                ]
            )
            self.get_bin(name=bin_id, url=url, copypaths=copypaths)

    def get_bin(self, name: str, url: str, copypaths: Dict[str, str] = None):
        """Get a remote bin package, and put any bin contents into a bin path.

        Make sure that we have the bin in scope; download it if we don't

        """
        path = shutil.which(name, os.X_OK)

        if path is None:
            logger.info("bin-helper can't find %s. Downloading it from %s", name, url)

            # First get the url
            with requests.get(url, allow_redirects=True) as res:
                res.raise_for_status()

                # try to decide on a name for the download
                file_name = None
                content_disposition = res.headers.get("content-disposition")
                if content_disposition is not None:
                    file_name_matches = re.findall("filename=(.+)", content_disposition)
                    if file_name_matches:
                        file_name = file_name_matches[0]
                if file_name is None:
                    # @NOTE this may not be a reliable way of getting a name from a url
                    file_name = os.path.basename(url)

                # write the file into our local path
                local_file = os.path.join(self.local_path, file_name)
                with open(local_file, "wb") as fil:
                    fil.write(res.content)

            # downloaded url is a tarfile.  Copy only the files out of the tarfile
            # that are suggested by the "copy" part of the config, and make sure
            # to 'chmod a+x' them.
            if tarfile.is_tarfile(local_file):
                _untar(local_file, self.local_path, copypaths)
                os.remove(local_file)

            # downloaded url is a zipfile.  Copy only the files out of the zipfile
            # that are suggested by the "copy" part of the config, and make sure
            # to 'chmod a+x' them.
            elif zipfile.is_zipfile(local_file):
                _unzip(local_file, self.local_path, copypaths)
                os.remove(local_file)

            # this case doesn't really make sense, but is possible based on config
            # where you specified a download url and multiple copy paths.
            # I don't think it has any actual usecases, but I wrote it anyway.
            elif copypaths:
                _copyfiles(self.local_path, copypaths)

            # In this case, you specified no `copy` items, so your URL must be
            # a single file that we downloaded.  Let's just make sure it is
            # executable.
            elif os.path.isfile(local_file):
                os.chmod(local_file, os.stat(local_file).st_mode | stat.S_IEXEC)


def _untar(local_file, local_path, copypaths):
    """Untar a file.

    Extract, copy and chmod executables out of a tar file.

    """
    with tarfile.open(local_file) as taf:
        members = taf.getmembers()
        for (bin_name, from_name) in copypaths.items():
            for member in members:
                if member.name == from_name:
                    member.name = bin_name
                    taf.extract(member, local_path)
                    bin_path = os.path.join(local_path, bin_name)
                    os.chmod(bin_path, os.stat(bin_path).st_mode | stat.S_IEXEC)
                    break


def _unzip(local_file, local_path, copypaths):
    """Unzip a file.

    Extract, copy and chmod executables out of a zip file.

    """
    with zipfile.ZipFile(local_file) as zif:
        members = zif.infolist()
        for (bin_name, from_name) in copypaths.items():
            for member in members:
                if member.filename == from_name:
                    bin_path = os.path.join(local_path, bin_name)
                    with zif.open(member) as source, open(bin_path, "wb") as target:
                        shutil.copyfileobj(source, target)
                    os.chmod(bin_path, os.stat(bin_path).st_mode | stat.S_IEXEC)


def _copyfiles(local_path, copypaths):
    """Copy files from one path to another.

    Just move and chmod files from one path to another, and chmod executables.

    """
    for (bin_name, from_name) in copypaths.items():
        bin_path = os.path.join(local_path, bin_name)
        from_path = os.path.join(local_path, from_name)
        os.rename(from_path, bin_path)
        os.chmod(bin_path, os.stat(bin_path).st_mode | stat.S_IEXEC)



"""

Metta docker CLI plugin.

Provides functionality to manually interact with the various docker plugins.

"""
import logging
from typing import Dict, Any

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import Fixtures
from mirantis.testing.metta_health.healthcheck import Health
from mirantis.testing.metta_cli.base import CliBase, cli_output

from .client import METTA_PLUGIN_ID_DOCKER_CLIENT

logger = logging.getLogger("metta.cli.docker")

METTA_PLUGIN_ID_DOCKER_CLI = "metta_docker_cli"
""" metta plugin_id for the launchpad cli plugin """


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class DockerCliPlugin(CliBase):
    """Fire command/group generator for various docker plugin commands."""

    def fire(self) -> Dict[str, Any]:
        """Return command groups for Docker plugins."""
        if (
            self._environment.fixtures().get(
                plugin_id=METTA_PLUGIN_ID_DOCKER_CLIENT,
                exception_if_missing=False,
            )
            is not None
        ):

            return {"docker": DockerClientGroup(self._environment)}

        return {}


class DockerClientGroup:
    """Base Fire command group for terraform client cli commands."""

    def __init__(self, environment: Environment):
        """Add additional command groups for plugins and inject environment."""
        self._environment: Environment = environment

    def _select_client(self, instance_id: str = "") -> Fixtures:
        """Pick a matching client."""
        if instance_id:
            return self._environment.fixtures().get(
                plugin_id=METTA_PLUGIN_ID_DOCKER_CLIENT,
                instance_id=instance_id,
            )

        # Get the highest priority workload
        return self._environment.fixtures().get(
            plugin_id=METTA_PLUGIN_ID_DOCKER_CLIENT,
        )

    # deep argument is an info() standard across plugins
    # pylint: disable=unused-argument
    def info(self, workload: str = "", deep: bool = False):
        """Get info about a client plugin."""
        docker_fixture = self._select_client(instance_id=workload)
        return cli_output(docker_fixture.info())

    def health(self, client: str = ""):
        """Run healthcheck."""
        fixture = self._select_client(instance_id=client)
        plugin = fixture.plugin
        health: Health = plugin.health()
        return cli_output(
            {
                "fixture": {
                    "plugin_id": fixture.plugin_id,
                    "instance_id": fixture.instance_id,
                    "priority": fixture.priority,
                },
                "status": health.status(),
                "messages": list(health.messages()),
            }
        )

    def swarm_info(self, client: str = ""):
        """List all containers on the docker client."""
        docker_client = self._select_client(instance_id=client).plugin
        return cli_output(
            {
                "version": docker_client.swarm.version,
                "attrs": docker_client.swarm.attrs,
            }
        )

    def nodes_list(self, client: str = ""):
        """List all nodes in the docker swarm."""
        docker_client = self._select_client(instance_id=client).plugin
        return cli_output(
            list(
                {
                    "id": node.id,
                    "short_id": node.short_id,
                    "version": node.version,
                    "attrs": node.attrs,
                }
                for node in docker_client.nodes.list()
            )
        )

    # 'all' is effectively a passthrough variable
    # pylint: disable=redefined-builtin
    def container_list(self, client: str = "", all: bool = True):
        """List all containers on the docker client."""
        docker_client = self._select_client(instance_id=client).plugin
        container_info = []
        for container in docker_client.containers.list(all=all):
            container_info.append(
                {
                    "name": container.name,
                    "id": container.id,
                    "image": container.image.tags,
                    "status": container.status,
                }
            )

        return cli_output(container_info)

    def run_hello_world(self, client: str = ""):
        """Run the hello-world image."""
        docker_client = self._select_client(instance_id=client).plugin
        return cli_output(docker_client.containers.run(image="hello-world", remove=True))



"""

Metta sonobuoy CLI plugin.

Provides functionality to manually run and inspect sonobuoy jobs

"""
import logging
import time
from typing import Dict, Any

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import Fixture
from mirantis.testing.metta_cli.base import CliBase, cli_output

from .client import METTA_SONOBUOY_CLIENT_PLUGIN_ID, SonobuoyClientPlugin
from .workload import METTA_SONOBUOY_WORKLOAD_PLUGIN_ID
from .results import Status

logger = logging.getLogger("metta.cli.sonobuoy")

METTA_PLUGIN_ID_SONOBUOY_CLI = "metta_sonobuoy_cli"
""" cli plugin_id for the sonobuoy plugin """


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class SonobuoyCliPlugin(CliBase):
    """Fire command/group generator for sonobuoy commands."""

    def fire(self):
        """Return CLI Command group."""
        commands: Dict[str, Any] = {}

        if (
            self._environment.fixtures().get(
                plugin_id=METTA_SONOBUOY_CLIENT_PLUGIN_ID,
                exception_if_missing=False,
            )
            is not None
        ):
            commands["sonobuoy"] = SonobuoyClientGroup(self._environment)

        return commands


class SonobuoyClientGroup:
    """Sonobuoy cli commands."""

    def __init__(self, environment: Environment):
        """Inject Environment into command group."""
        self._environment: Environment = environment

        if (
            self._environment.fixtures().get(
                plugin_id=METTA_SONOBUOY_WORKLOAD_PLUGIN_ID,
                exception_if_missing=False,
            )
            is not None
        ):
            self.workload = SonobuoyWorkloadGroup(self._environment)

    def _select_fixture(self, instance_id: str = "") -> Fixture:
        """Pick a matching client plugin."""
        try:
            if instance_id:
                return self._environment.fixtures().get(
                    plugin_id=METTA_SONOBUOY_CLIENT_PLUGIN_ID,
                    instance_id=instance_id,
                )

            # Get the highest priority provisioner
            return self._environment.fixtures().get(
                plugin_id=METTA_SONOBUOY_CLIENT_PLUGIN_ID,
            )

        except KeyError as err:
            raise ValueError("No usable sonobuoy client was found.") from err

    def info(self, instance_id: str = "", deep: bool = False):
        """Get info about a sonobuoy plugin."""
        fixture = self._select_fixture(instance_id=instance_id)
        return cli_output(fixture.info(deep=deep))

    def status(self, instance_id: str = ""):
        """Get active sonobuoy status."""
        client_plugin = self._select_fixture(instance_id=instance_id).plugin
        status = client_plugin.status()

        if status is None:
            status_info = {"status": "None", "plugins": {}}
        else:
            status_info = {
                "status": status.status,
                "plugins": {
                    plugin_id: status.plugin(plugin_id) for plugin_id in status.plugin_list()
                },
            }

        return cli_output(status_info)

    # pylint: disable=protected-access
    def crb(self, instance_id: str = "", remove: bool = False):
        """Create the crb needed to run sonobuoy."""
        client_plugin = self._select_fixture(instance_id=instance_id).plugin

        if remove:
            client_plugin._sonobuoy._delete_k8s_crb()
        else:
            client_plugin._sonobuoy._create_k8s_crb()

    def run(self, instance_id: str = "", wait: bool = False):
        """Run sonobuoy."""
        client_plugin = self._select_fixture(instance_id=instance_id).plugin
        client_plugin.run(wait=wait)

    def wait(self, instance_id: str = "", step: int = 5, limit: int = 1000):
        """Wait until no longer running."""
        client_plugin = self._select_fixture(instance_id=instance_id).plugin
        print("{")
        for i in range(0, limit):
            status = client_plugin.status()
            if status is None:
                status_info = {"status": "None", "plugins": {}}
            else:
                status_info = {
                    "status": status.status.value,
                    "plugins": {
                        plugin_id: status.plugin(plugin_id) for plugin_id in status.plugin_list()
                    },
                }

            print(f"{i}: {status_info},")
            if status is None or status.status not in [Status.RUNNING]:
                break

            time.sleep(step)
        print("}")

    def delete(self, instance_id: str = "", wait: bool = False):
        """Remove all sonobuoy infrastructure."""
        client_plugin = self._select_fixture(instance_id=instance_id).plugin
        client_plugin.delete(wait=wait)

    def logs(self, instance_id: str = "", follow: bool = False):
        """Retrieve sonobuoy logs."""
        client_plugin = self._select_fixture(instance_id=instance_id).plugin
        client_plugin.logs(follow=follow)

    def results(self, instance_id: str = ""):
        """Retrieve sonobuoy results."""
        client_plugin = self._select_fixture(instance_id=instance_id).plugin
        results: str = client_plugin.results()
        return results

    def retrieve(self, instance_id: str = ""):
        """Retrieve the results from the sonobuoy workload instance."""
        client_plugin = self._select_fixture(instance_id=instance_id).plugin

        try:
            client_plugin.retrieve()
        # broad catch to allow message formatting for cli output
        # pylint: disable=broad-except
        except Exception as err:
            logger.error("Retrieve failed: %s", err)

    def version(self, instance_id: str = ""):
        """Retrieve the sonobuoy version string."""
        client_plugin = self._select_fixture(instance_id=instance_id).plugin
        return cli_output(client_plugin.version())

    def health(self, instance_id: str = ""):
        """Retrieve the results from the sonobuoy workload instance."""
        client_plugin = self._select_fixture(instance_id=instance_id).plugin
        health = client_plugin.health()

        return cli_output(
            {
                "status": health.status(),
                "messages": list(health.messages()),
            }
        )

    def create_k8s_crb(self, instance_id: str = ""):
        """Create the cluster role binding that sonobuoy needs."""
        client_plugin = self._select_fixture(instance_id=instance_id).plugin
        return cli_output(client_plugin._sonobuoy.create_k8s_crb())

    def delete_k8s_crb(self, instance_id: str = ""):
        """Remove the cluster role binding that we created."""
        client_plugin = self._select_fixture(instance_id=instance_id).plugin
        return cli_output(client_plugin._sonobuoy.delete_k8s_crb())


class SonobuoyWorkloadGroup:
    """Sonobuoy workload commands."""

    def __init__(self, environment: Environment):
        """Inject Environment into command group."""
        self._environment: Environment = environment

    def _select_fixture(self, instance_id: str = "") -> Fixture:
        """Pick a matching workload plugin."""
        try:
            if instance_id:
                return self._environment.fixtures().get(
                    plugin_id=METTA_SONOBUOY_WORKLOAD_PLUGIN_ID,
                    instance_id=instance_id,
                )

            # Get the highest priority provisioner
            return self._environment.fixtures().get(
                plugin_id=METTA_SONOBUOY_WORKLOAD_PLUGIN_ID,
            )

        except KeyError as err:
            raise ValueError("No usable sonobuoy client was found.") from err

    def _prepared_fixture(self, instance_id: str = "") -> Fixture:
        """Pick a matching workload plugin and prepare it."""
        sonobuoy_fixture: Fixture = self._select_fixture(instance_id=instance_id)
        sonobuoy_fixture.plugin.prepare()
        return sonobuoy_fixture

    def _client_plugin(self, instance_id: str = "") -> SonobuoyClientPlugin:
        """Get the workload's client plugin."""
        return self._prepared_fixture(instance_id=instance_id).plugin.get_client_plugin()

    def info(self, instance_id: str = "", deep: bool = False):
        """Get info about a sonobuoy plugin."""
        fixture = self._select_fixture(instance_id=instance_id)
        return cli_output(fixture.info(deep=deep))

    def apply(self, instance_id: str = "", wait: bool = False):
        """Run sonobuoy."""
        workload_plugin = self._select_fixture(instance_id=instance_id).plugin
        workload_plugin.apply(wait=wait)

    def destroy(self, instance_id: str = "", wait: bool = False):
        """Remove all sonobuoy infrastructure."""
        workload_plugin = self._select_fixture(instance_id=instance_id).plugin
        workload_plugin.destroy(wait=wait)

    def logs(self, instance_id: str = "", follow: bool = False):
        """Retrieve sonobuoy logs."""
        client_plugin = self._client_plugin(instance_id=instance_id)
        client_plugin.logs(follow=follow)

    def retrieve(self, instance_id: str = ""):
        """Retrieve the results from the sonobuoy workload instance."""
        client_plugin = self._client_plugin(instance_id=instance_id)

        try:
            client_plugin.retrieve()
        # broad catch to allow message formatting for cli output
        # pylint: disable=broad-except
        except Exception as err:
            logger.error("Retrieve failed: %s", err)



"""

Metta client plugin for terraform.

The metta client plugin is a client which can run terraform using
an assigned chart.

"""
import logging
from typing import Dict, Any
import os
import json

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixture import Fixtures
from mirantis.testing.metta.output import METTA_PLUGIN_INTERFACE_ROLE_OUTPUT
from mirantis.testing.metta_common import (
    METTA_PLUGIN_ID_OUTPUT_DICT,
    METTA_PLUGIN_ID_OUTPUT_TEXT,
)

from .terraform import TerraformClient


logger = logging.getLogger("metta_terraform:client")


METTA_TERRAFORM_CLIENT_PLUGIN_ID = "metta_terraform_client"
""" Terraform provisioner plugin id """


class TerraformClientPlugin:
    """Metta terraform client."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        environment: Environment,
        instance_id: str,
        chart_path: str,
        state_path: str,
        tfvars: Dict[str, Any],
        tfvars_path: str,
    ):
        """Initial client configuration."""
        self._environment: Environment = environment
        """ Environemnt in which this plugin exists """
        self._instance_id: str = instance_id
        """ Unique id for this plugin instance """

        self.tfvars: Dict[str, Any] = tfvars
        """Terraform vars to pass to TF as a tfvars file."""
        self._tfvars_path: str = os.path.realpath(tfvars_path)
        """Path to write for the tfvars file."""

        self._tf_handler = TerraformClient(
            working_dir=os.path.realpath(chart_path),
            state_path=os.path.realpath(state_path),
            tfvars_path=os.path.realpath(tfvars_path),
        )
        """Terraform handler which actually runs terraform commands."""

        self.fixtures = Fixtures()
        """All fixtures added to this plugin, which are primarily TF output plugins."""

        # if the cluster is already provisioned then we can get outputs from it
        try:
            self.make_fixtures()
        # pylint: disable=broad-except
        except Exception:
            pass

    # deep argument is an info() standard across plugins
    # pylint: disable=unused-argument
    def info(self, deep: bool = False):
        """Get info about the client plugin.

        Returns:
        --------
        Dict of keyed introspective information about the plugin.

        """
        info = {
            "config": {
                "tfvars": self.tfvars,
                "tfvars_path": self._tfvars_path,
            },
            "client": self._tf_handler.info(deep=deep),
        }

        return info

    def state(self):
        """Return the terraform state contents."""
        return self._tf_handler.state()

    def init(self):
        """Run terraform init."""
        self._tf_handler.init()

    def apply(self, lock: bool = True):
        """Apply a terraform plan."""
        self._make_tfvars_file()
        self._tf_handler.apply(lock=lock)
        self.make_fixtures()

    def destroy(self, lock: bool = True):
        """Apply a terraform plan."""
        self._tf_handler.destroy(lock=lock)
        self._rm_tfvars_file()

    def test(self):
        """Apply a terraform plan."""
        self._make_tfvars_file()
        return self._tf_handler.test()

    def plan(self):
        """Check a terraform plan."""
        self._make_tfvars_file()
        return self._tf_handler.plan()

    def output(self, name: str = ""):
        """Retrieve terraform outputs.

        Run the terraform output command, to retrieve outputs.
        Outputs are returned always as json as it is the only way to machine
        parse outputs properly.

        Returns:
        --------
        If you provided a name, then a single output is returned, otherwise a
        dict of outputs is returned.

        """
        return self._tf_handler.output(name=name)

    def _make_tfvars_file(self):
        """Write the vars file."""
        os.makedirs(os.path.dirname(self._tfvars_path), exist_ok=True)
        with open(self._tfvars_path, "w", encoding="utf8") as var_file:
            json.dump(self.tfvars, var_file, sort_keys=True, indent=4)

    def _rm_tfvars_file(self):
        """Remove any created vars file."""
        tfvars_path = self._tfvars_path
        if os.path.isfile(tfvars_path):
            os.remove(tfvars_path)

    def make_fixtures(self):
        """Retrieve an output from terraform.

        For other METTA plugins we can just load configuration, and creating
        output plugin instances from various value in config.

        We do that here, but  we also want to check of any outputs exported by
        the terraform root module, which we get using the tf client.

        If we find a root module output without a matching config output
        defintition then we make some assumptions about plugin type and add it
        to the list. We make some simple investigation into output plugin types
        and pick either the contrib.common.dict or contrib.common.text plugins.

        If we find a root module output that matches an output that was
        declared in config then we use that.  This allows config to define a
        plugin_id which will then be populated automatically.  If you know what
        type of data you are expecting from a particular tf output then you can
        prepare and config for it to do things like setting default values.

        Priorities can be used in the config.

        """
        # now we ask TF what output it nows about and merge together those as
        # new output plugins.
        # tf.outputs() produces a list of (sensitive:bool, type: [str,  object,
        # value:Any])
        for output_key, output_struct in self.output().items():
            # Here is the kind of info we can get out of terraform
            # output_sensitive = bool(output_struct['sensitive'])
            # """ Whether or not the output contains sensitive data """
            output_type = output_struct["type"][0]
            # output_spec = output_struct['type'][1]
            # """ A structured spec for the type """
            output_value = output_struct["value"]

            # see if we already have an output plugin for this name
            fixture = self.fixtures.get(
                interfaces=[METTA_PLUGIN_INTERFACE_ROLE_OUTPUT],
                instance_id=output_key,
                exception_if_missing=False,
            )
            if fixture is not None:
                if hasattr(fixture.plugin, "set_data"):
                    fixture.plugin.set_data(output_value)
                elif hasattr(fixture.plugin, "set_text"):
                    fixture.plugin.set_text(str(output_value))

            else:
                # we only know how to create 2 kinds of outputs
                if output_type == "object":
                    fixture = self._environment.new_fixture(
                        plugin_id=METTA_PLUGIN_ID_OUTPUT_DICT,
                        instance_id=output_key,
                        priority=self._environment.plugin_priority(delta=5),
                        arguments={"data": output_value},
                        labels={
                            "parent_plugin_id": METTA_TERRAFORM_CLIENT_PLUGIN_ID,
                            "parent_instance_id": self._instance_id,
                        },
                    )
                else:
                    fixture = self._environment.new_fixture(
                        plugin_id=METTA_PLUGIN_ID_OUTPUT_TEXT,
                        instance_id=output_key,
                        priority=self._environment.plugin_priority(delta=5),
                        arguments={"text": str(output_value)},
                        labels={
                            "parent_plugin_id": METTA_TERRAFORM_CLIENT_PLUGIN_ID,
                            "parent_instance_id": self._instance_id,
                        },
                    )

                self.fixtures.add(fixture)



"""

Utility mixins for Environment objects.

"""
from logging import getLogger
from typing import List, Dict, Any, Callable, Optional, Union
import random
import string

from configerus.config import Config
from configerus.loaded import Loaded, LOADED_KEY_ROOT
from configerus.validator import ValidationError
from configerus.contrib.jsonschema import PLUGIN_ID_VALIDATE_JSONSCHEMA

from .plugin import (
    METTA_PLUGIN_CONFIG_KEY_PLUGINID,
    METTA_PLUGIN_CONFIG_KEY_INSTANCEID,
    METTA_PLUGIN_CONFIG_KEY_ARGUMENTS,
    METTA_PLUGIN_CONFIG_KEY_PLUGINLABELS,
)
from .fixture import (
    Fixture,
    Fixtures,
    METTA_FIXTURE_CONFIG_KEY_PRIORITY,
    METTA_FIXTURES_CONFIG_FIXTURES_LABEL,
    METTA_FIXTURE_VALIDATION_JSONSCHEMA,
)

logger = getLogger("metta.env.mixins")

DEFAULT_FIXTURE_PRIORITY: int = 60
"""Default priority for plugin fixtures if not specified."""

METTA_PLUGIN_CONFIG_KEY_FROM_CONFIG = "from_config"
""" config key that indicates that the plugin will be build from aconfig label/key pair """


class FixtureBuildingFromConfigMixin:
    """Functions for building fixtures from config.

    This class contains various functions that can be used to build Fixtures
    by interpreting either Configerus Config data or raw Python primitives.
    The class can be used as a Mixin to add the standardized building approach
    to any object that can provide a Config source and a Fixture building
    callback.

    The provided Config object is used as an optional source of data for
    building, while the callback is passed raw Fixture building arguments as
    interpreted. This allows avoiding being too strict about how Fixtures are
    finally built and we can stay out of some ugly circular imports.

    """

    def __init__(
        self,
        config: Config,
        builder_callback: Callable[
            [str, str, int, Optional[Dict[str, Any]], Optional[Dict[str, Any]], Optional[bool]],
            Fixture,
        ],
        default_priority: int = DEFAULT_FIXTURE_PRIORITY,
    ):
        """Capture an environment used for constructor args for fixtures.

        Parameters:
        -----------
        config (Config) : A Configerus Config object which will be used as a
            source of information for building Fixtures by interpreting config
            as the parametrization for Fixture arguments.

        builder_callback (Callable) : A call back function which is meant to
            be used to actually build Fixtures from processed config.

            plugin_id (str) : METTA plugin id to tell us what plugin factory to use;

                @see .plugin.Factory for more details on how plugins are loaded.

            instance_id (str) : string instance id that will be passed to the new
                plugin object;

            priority (int) : Integer priority 1-100 for comparative prioritization
                between other plugins;

            arguments (Dict[str, Any]) : Keyword Arguments which should be passed to
                the plugin constructor after environment and instance_id;

            labels (Dict[str, str]) : Keyword/value labels which are to be associated
                with the plugin/fixture;

            replace_existing (bool) : Replace any existing matching fixture.

            e.g.

                def callback(
                        plugin_id: str,
                        instance_id: str,
                        priority: int,
                        arguments: Dict[str, Any] = None,
                        labels: Dict[str, Any] = None,
                        replace_existing=False,
                )


        """
        self._config: Config = config
        """Configerus Config object, for searching for fixture settings."""
        self.builder_callback: Callable = builder_callback
        """Callable callback ffixture factory used when settings are found."""
        self._default_priority: int = default_priority
        """Keep the default priority."""

    def plugin_priority(self, delta: int = 0):
        """Return a default plugin priority with a delta.

        Not sure what priority to give a plugin, or maybe you want a priority
        relative to a median; use this.

        Returns:
        --------
        Integer default priority with the provided delta added.

        """
        return self._default_priority + delta

    # Generic Plugin construction

    # pylint: disable=too-many-arguments
    # This is what it takes to build a plugin.
    def add_fixtures_from_config(
        self,
        label: str = METTA_FIXTURES_CONFIG_FIXTURES_LABEL,
        base: Union[str, List[Any]] = LOADED_KEY_ROOT,
        validator: str = "",
        exception_if_missing: bool = False,
        arguments: Dict[str, Any] = None,
        labels: Dict[str, str] = None,
    ) -> Fixtures:
        """Create plugin fixtures from some config.

        This method will interpret some config values as being usable to build a
        collection of plugins from.  The plugins will be built, wrapped as
        fixtures and added to the Environment.  The plugins will then be
        returned as a Fixtures collection.

        Parameters:
        -----------
        config (Config) : Used to load and get the plugin configuration

        label (str) : config label to load to pull plugin configuration. That
            label is loaded and config is pulled to produce a list of plugins

        base (str|List) : config key to get Dict of plugins configurations. This
            should point to a dict of plugin configurations.
            A list of strings is valid as configerus.loaded.get() can take that
            as an argument.
            We call this base instead of key as we will be searching for
            sub-paths to pull individual elements

        validator (str) : optionally use a configerus validator on the instance
            config/dict before a plugin is created.

        arguments (Dict[str, Any]) : A Dict of named arguments to pass to the
            plugin constructor.  Constructors should be able to work without
            requiring the arguments, but these tend to be pivotal for them.

        labels (Dict[str, str]) : Dictionary of labels which should be added to
            the created fixture.

        Returns:
        --------
        Fixtures set as directed by the passed config.

        Raises:
        -------
        If you ask for a plugin which has not been registered, then you're going
        to get a NotImplementedError exception.
        To make sure that your desired plugin is registered, make sure to import
        the module that contains the factory method with a decorator.

        """
        if arguments is None:
            arguments = {}

        # plugin set which will be used to create new plugins
        fixtures: Fixtures = Fixtures()

        try:
            plugin_config = self._config.load(label)

            # check to see if we are being directed elsewhere for config
            # using a "from_config" directive
            try:
                config_fromconfig = plugin_config.get([base, METTA_PLUGIN_CONFIG_KEY_FROM_CONFIG])

                if isinstance(config_fromconfig, dict):
                    logger.debug("Using from_config, and passing label/base as arguments")
                    label = plugin_config.get(
                        [base, METTA_PLUGIN_CONFIG_KEY_FROM_CONFIG, "label"], default=label
                    )
                    base = plugin_config.get(
                        [base, METTA_PLUGIN_CONFIG_KEY_FROM_CONFIG, "base"],
                        default=LOADED_KEY_ROOT,
                    )

                    plugin_config = self._config.load(label)
                else:
                    logger.debug("Using from_config, and passing current label/base as arguments")
            except KeyError:
                pass
            plugin_list = plugin_config.get(base)

        except KeyError as err:
            if exception_if_missing:
                raise KeyError("Could not load any config for plugin generation") from err
            return fixtures

        for instance_id in plugin_list.keys():
            # We only accept string instance ids
            instance_id = str(instance_id)

            # This fixture gets effectively added to 2 different Fixtures object.
            # 1. we manually add it to our Fixtures object for this function call
            # 2. the add_fixture_from_config() adds it to the fixture for this
            #    environment object.
            fixture = self.add_fixture_from_config(
                label=label,
                base=[base, instance_id] if base else instance_id,
                instance_id=instance_id,
                labels=labels,
                validator=validator,
                arguments=arguments,
            )
            fixtures.add(fixture)

        return fixtures

    # pylint: disable=too-many-arguments
    # This is what it takes to build a plugin.
    def add_fixture_from_config(
        self,
        label: str,
        base: Union[str, List[Any]] = LOADED_KEY_ROOT,
        instance_id: str = "",
        priority: int = -1,
        validator: str = "",
        arguments: Dict[str, Any] = None,
        labels: Dict[str, str] = None,
    ) -> Fixture:
        """Create and add a new plugin fixture from some config.

        This method will interpret some config values as being usable to build a
        single plugin.  The plugin will be built, wrapped as a Fixture, and
        added to the environment. The plugin Fixture is returned.

        @see add_fixture_from_loadedconfig

        @note we could validate here instead of passing on the validator, but we
            pass it on to keep parity across a number of cases, and to keep as
            much complex logic in one place.

        Parameters:
        -----------
        config (Config) : Used to load and get the plugin configuration

        label (str) : config label to load to pull plugin configuration. That
            label is loaded and config is pulled to produce a list of plugins.

        base (str|List) : config key used as a .get() base for all gets.  With
            this you can instruct to pull config from a section of loaded
            config.
            A list of strings is valid because configerus.loaded.get() can take
            that as an argument. We will be using the list syntax anyway. We
            call this base instead of key as we will be searching for  sub-paths
            to pull individual elements.

        instance_id (str) : optionally pass an instance_id for the item.

        labels (Dict[str, str]) : Dictionary of labels which should be added to
            the created fixture.

        validator (str) : optionally use a configerus validator on the entire
            .get() for the instance config.

        Returns:
        --------
        A Fixture object that wraps the created plugin.

        The Fixtures has already been added to the environment, but is returned
        so that the consumer can act on it separately without haveing to
        search for it.

        Raises:
        -------
        If you ask for a plugin which has not been registered, then you're going
        to get a NotImplementedError exception. To make sure that your desired
        plugin is registered, make sure to import the module that contains the
        factory method with a decorator.

        A ValueError is thrown if the plugin cannot be created due to missing
        plugin configuration/argument values.  This means that we could not
        determine how to create the plugin.

        A configerus.validate.ValidationError will be raised if a validation
        target was passed and validation failed.

        """
        if arguments is None:
            arguments = {}

        try:
            # loaded configuration for the plugin
            plugin_loaded = self._config.load(label)
        except KeyError as err:
            raise KeyError(f"Could not load plugin config source {label}") from err

        # if no arguments were specified, then consider a special case of
        # 'from_config' which build arguments for a config based plugin
        # which will take a config label/base-key pair as config arguments
        #
        # There is a special case where if the passed from_config is not a
        # dict then the same config label/base received is used.
        try:
            config_fromconfig = plugin_loaded.get([base, METTA_PLUGIN_CONFIG_KEY_FROM_CONFIG])

            if isinstance(config_fromconfig, dict):
                logger.debug("Using from_config, and passing label/base as arguments")
                arguments = config_fromconfig
            else:
                logger.debug("Using from_config, and passing current label/base as arguments")
                arguments = {"label": label, "base": base}
        except KeyError:
            pass

        return self.add_fixture_from_loadedconfig(
            loaded=plugin_loaded,
            base=base,
            instance_id=instance_id,
            priority=priority,
            validator=validator,
            arguments=arguments,
            labels=labels,
        )

    # This is where we centralize all logic around creating fixtures, so it is complex
    # pylint: disable=too-many-branches, too-many-locals, too-many-arguments
    def add_fixture_from_loadedconfig(
        self,
        loaded: Loaded,
        base: Union[str, List[Any]] = LOADED_KEY_ROOT,
        instance_id: str = "",
        priority: int = -1,
        labels: Dict[str, str] = None,
        validator: str = "",
        arguments: Dict[str, Any] = None,
    ) -> Fixture:
        """Create a plugin from a Configerus loaded config object.

        This method will interpret some config values as being usable to build
        plugin. This function starts with a loaded config object because we can
        leverage that from more starting points.

        Using a configerus config object allows us to leverage advanced
        configerus features such as tree searching, formatting and validation.

        What is looked for:

        1. valdiators if we need to validate the entire label/key before using
        2. plugin_id : which will tell us what plugin to load
        3. optional instance_id if none was passed
        4. config if you want config added - ONLY if fixtures is None
           (plugins in Fixtures cannot override config objects)
        5. arguments that will be executed on an argument() method if the
            plugin has it.

        @TODO we should probably allow a setting to allow replacing existing
            fixtures in the environment.

        Parameters:
        -----------
        config (Config) : Used to load and get the plugin configuration

        label (str) : config label to load to pull plugin configuration. That
            label is loaded and config is pulled to produce a list of plugins.

        base (str|List) : config key used as a .get() base for all gets.  With
            this you can instruct to pull config from a section of loaded
            config.
            A list of strings is valid because configerus.loaded.get() can take
            that as an argument. We will be using the list syntax anyway.
            We call this base instead of key as we will be searching for
            sub-paths to pull individual elements.

        instance_id (str) : optionally pass an instance_id for the item.

        labels (Dict[str, str]) : Dictionary of labels which should be added to
            the created fixture.

        arguments (Dict[str, str]) : Dictionary of args which should be passed
            to the plugin constructor.

        validator (str) : optionally use a configerus validator on the entire
            .get() for the instance config.

        Returns:
        --------
        A Fixture object with the new plugin added

        The Fixtures has already been added to the environment, but is returned
        so that the consumer can act on it separately without haveing to
        search for it.

        Raises:
        -------
        If you ask for a plugin which has not been registered, then you're going
        to get a NotImplementedError exception. To make sure that your desired
        plugin is registered, make sure to import the module that contains the
        factory method with a decorator.

        A ValueError is thrown if the plugin cannot be created due to missing
        plugin configuration/argument values.  This means that we could not
        determine how to create the plugin.

        A configerus.validate.ValidationError will be raised if a validation
        target was passed and validation failed.

        """
        # Retrieve all of the plugin config, to test that it exists
        # it might be expensive to retrieve all this but we do it to catch early
        # faults in config.
        if not loaded.has(base):
            raise ValueError("Cannot build plugin as provided config was empty.")

        # get any validators from config, defaulting to just the jsonschema
        # validator for a fixture
        validators = loaded.get(
            [base, METTA_FIXTURE_VALIDATION_JSONSCHEMA],
            default=[{PLUGIN_ID_VALIDATE_JSONSCHEMA: METTA_FIXTURE_VALIDATION_JSONSCHEMA}],
        )

        if validator:
            # if a validator arg was passed in, then add it
            validators.append(validator)
        if len(validators):
            # Run configerus validation on the config base once per validator
            try:
                for val in validators:
                    loaded.get(base, validator=val)
            except ValidationError as err:
                raise err

        try:
            plugin_id = str(loaded.get([base, METTA_PLUGIN_CONFIG_KEY_PLUGINID]))
        except KeyError as err:
            full_config = loaded.get(base)
            raise ValueError(
                "Could not find a plugin_id when trying to create a "
                f"plugin from config: {full_config}"
            ) from err

        # if no instance_id was passed, try to load one or just make one up
        if not instance_id:
            instance_id = str(loaded.get([base, METTA_PLUGIN_CONFIG_KEY_INSTANCEID], default=""))
            if not instance_id:
                instance_rand = "".join(random.choice(string.ascii_lowercase) for i in range(10))
                instance_id = f"{plugin_id}-{instance_rand}"

        if priority < 0:
            # instance priority - this is actually a stupid way to get it
            priority = int(
                loaded.get(
                    [base, METTA_FIXTURE_CONFIG_KEY_PRIORITY],
                    default=self.plugin_priority(),
                )
            )

        config_arguments = loaded.get([base, METTA_PLUGIN_CONFIG_KEY_ARGUMENTS], default={})
        if len(config_arguments) > 0:
            if arguments is None:
                arguments = {}
            else:
                # if we have config arguments from two different sources, then
                # add the config args to a copy of the function parameter args.
                # We use a copy so that we make no context mistakes by altering
                # a passed Dict that may get used for more than one plugin.
                arguments = arguments.copy()

            arguments.update(config_arguments)

        config_labels = loaded.get([base, METTA_PLUGIN_CONFIG_KEY_PLUGINLABELS], default={})
        if len(config_labels) > 0:
            if labels is None:
                labels = {}
            else:
                labels = labels.copy()

            labels.update(config_labels)

        # Use the factory to make the .fixture.Fixture
        return self.builder_callback(
            plugin_id=plugin_id,
            instance_id=instance_id,
            priority=priority,
            arguments=arguments,
            labels=labels,
        )


class FixtureBuildingFromDictMixin:
    """Functions for building fixtures from Dicts.

    This class contains various functions that can be used to build Fixtures
    by interpreting either Configerus Config data or raw Python primitives.
    The class can be used as a Mixin to add the standardized building approach
    to any object that can provide Dict sources and a Fixture building
    callback.

    The callback is passed raw Fixture building arguments as
    interpreted. This allows avoiding being too strict about how Fixtures are
    finally built and we can stay out of some ugly circular imports.

    """

    def __init__(
        self,
        builder_callback: Callable[[str, str, int, Dict[str, Any], Dict[str, Any], bool], Fixture],
    ):
        """Capture an environment used for constructor args for fixtures.

        Parameters:
        -----------
        builder_callback (Callable) : A call back function which is meant to
            be used to actually build Fixtures from processed config.

            plugin_id (str) : METTA plugin id to tell us what plugin factory to use;

                @see .plugin.Factory for more details on how plugins are loaded.

            instance_id (str) : string instance id that will be passed to the new
                plugin object;

            priority (int) : Integer priority 1-100 for comparative prioritization
                between other plugins;

            arguments (Dict[str, Any]) : Keyword Arguments which should be passed to
                the plugin constructor after environment and instance_id;

            labels (Dict[str, str]) : Keyword/value labels which are to be associated
                with the plugin/fixture;

            replace_existing (bool) : Replace any existing matching fixture.

            e.g.

                def callback(
                        plugin_id: str,
                        instance_id: str,
                        priority: int,
                        arguments: Dict[str, Any] = None,
                        labels: Dict[str, Any] = None,
                        replace_existing=False,
                )


        """
        self.builder_callback: Callable = builder_callback
        """Callable callback ffixture factory used when settings are found."""

    # Generic Plugin construction

    def add_fixtures_from_dicts(
        self,
        plugins_list: List[Dict[str, Any]],
    ) -> Fixtures:
        """Create a Fixtures set if Fixtures from dict parameters."""
        fixtures: Fixtures = Fixtures()

        for plugin_dict in plugins_list:
            fixture = self.add_fixture_from_dict(plugin_dict)
            fixtures.add(fixture)

        return fixtures

    def add_fixture_from_dict(
        self,
        plugin_dict: Dict[str, Any],
    ) -> Fixture:
        """Create a single plugin from a Dict of information for it.

        Create a new plugin from a map/dict of settings for needed parameters.

        @see add_fixture_from_loadedconfig

        Parameters:
        -----------
        client_dict (Dict[str,Any]) : Dict from which all needed information will
            be pulled.  Optionally additional config sources can be included as
            well as arguments which could be passed to the plugin.

        Return:
        -------
        A Fixture object with the new plugin added

        The Fixtures has already been added to the environment, but is returned
        so that the consumer can act on it separately without haveing to
        search for it.

        """
        try:
            plugin_id = str(plugin_dict[METTA_PLUGIN_CONFIG_KEY_PLUGINID])
        except KeyError as err:
            raise ValueError(
                "Could not find a plugin_id when trying to create a "
                f"plugin from dict: {plugin_dict}"
            ) from err

        # if no instance_id was passed, try to load one or just make one up
        if METTA_PLUGIN_CONFIG_KEY_INSTANCEID in plugin_dict:
            instance_id = str(plugin_dict[METTA_PLUGIN_CONFIG_KEY_INSTANCEID])
        else:
            instance_rand = "".join(random.choice(string.ascii_lowercase) for i in range(10))
            instance_id = f"{plugin_id}-{instance_rand}"

        # instance priority - this is actually a stupid way to get it
        priority = 70
        if METTA_FIXTURE_CONFIG_KEY_PRIORITY in plugin_dict:
            priority = int(plugin_dict[METTA_FIXTURE_CONFIG_KEY_PRIORITY])

        arguments: Dict[str, Any] = {}
        if METTA_PLUGIN_CONFIG_KEY_ARGUMENTS in plugin_dict:
            arguments.update(plugin_dict[METTA_PLUGIN_CONFIG_KEY_ARGUMENTS])

        labels: Dict[str, Any] = {}
        if METTA_PLUGIN_CONFIG_KEY_PLUGINLABELS in plugin_dict:
            labels.update(plugin_dict[METTA_PLUGIN_CONFIG_KEY_PLUGINLABELS])

        # Use the factory to make the .fixture.Fixture
        fixture = self.builder_callback(
            plugin_id=plugin_id,
            instance_id=instance_id,
            priority=priority,
            arguments=arguments,
            labels=labels,
        )

        return fixture
